from __future__ import annotations

import typing as t

from sqlglot import tokenizer_core
from sqlglot.trie import new_trie

from sqlglot.tokenizer_core import Token, TokenizerCore, TokenType


# The sqlglotc distribution ships no importable `sqlglotc` module; it overlays
# compiled .so files onto sqlglot's modules, so detect it via the compiled core.
SQLGLOTC_INSTALLED = not getattr(tokenizer_core, "__file__", "py").endswith(".py")

try:
    import sqlglotrs  # type: ignore # noqa: F401

    if not SQLGLOTC_INSTALLED:
        import warnings

        warnings.warn(
            "sqlglot[rs] is deprecated and no longer compatible with sqlglot. "
            "Please use sqlglotc instead for faster parsing: pip install sqlglot[c]",
        )
except ImportError:
    pass

if t.TYPE_CHECKING:
    from sqlglot.dialects.dialect import DialectType


def _convert_quotes(arr: list[str | tuple[str, str]]) -> dict[str, str]:
    return dict((item, item) if isinstance(item, str) else (item[0], item[1]) for item in arr)


def _quotes_to_format(
    token_type: TokenType, arr: list[str | tuple[str, str]]
) -> dict[str, tuple[str, TokenType]]:
    return {k: (v, token_type) for k, v in _convert_quotes(arr).items()}


class _TokenizerBase:
    QUOTES: t.ClassVar[list[tuple[str, str] | str]]
    IDENTIFIERS: t.ClassVar[list[str | tuple[str, str]]]
    BIT_STRINGS: t.ClassVar[list[str | tuple[str, str]]]
    BYTE_STRINGS: t.ClassVar[list[str | tuple[str, str]]]
    HEX_STRINGS: t.ClassVar[list[str | tuple[str, str]]]
    RAW_STRINGS: t.ClassVar[list[str | tuple[str, str]]]
    HEREDOC_STRINGS: t.ClassVar[list[str | tuple[str, str]]]
    UNICODE_STRINGS: t.ClassVar[list[str | tuple[str, str]]]
    STRING_ESCAPES: t.ClassVar[list[str]]
    BYTE_STRING_ESCAPES: t.ClassVar[list[str]]
    ESCAPE_FOLLOW_CHARS: t.ClassVar[list[str]]
    IDENTIFIER_ESCAPES: t.ClassVar[list[str]]
    HINT_START: t.ClassVar[str]
    KEYWORDS: t.ClassVar[dict[str, TokenType]]
    SINGLE_TOKENS: t.ClassVar[dict[str, TokenType]]
    NUMERIC_LITERALS: t.ClassVar[dict[str, str]]
    VAR_SINGLE_TOKENS: t.ClassVar[set[str]]
    COMMANDS: t.ClassVar[set[TokenType]]
    COMMAND_PREFIX_TOKENS: t.ClassVar[set[TokenType]]
    HEREDOC_TAG_IS_IDENTIFIER: t.ClassVar[bool]
    STRING_ESCAPES_ALLOWED_IN_RAW_STRINGS: t.ClassVar[bool]
    NESTED_COMMENTS: t.ClassVar[bool]
    TOKENS_PRECEDING_HINT: t.ClassVar[set[TokenType]]
    HEREDOC_STRING_ALTERNATIVE: t.ClassVar[TokenType]
    COMMENTS: t.ClassVar[list[str | tuple[str, str]]]
    _QUOTES: t.ClassVar[dict[str, str]]
    _IDENTIFIERS: t.ClassVar[dict[str, str]]
    _FORMAT_STRINGS: t.ClassVar[dict[str, tuple[str, TokenType]]]
    _STRING_ESCAPES: t.ClassVar[set[str]]
    _BYTE_STRING_ESCAPES: t.ClassVar[set[str]]
    _ESCAPE_FOLLOW_CHARS: t.ClassVar[set[str]]
    _IDENTIFIER_ESCAPES: t.ClassVar[set[str]]
    _COMMENTS: t.ClassVar[dict[str, str | None]]
    _KEYWORD_TRIE: t.ClassVar[dict[str | int, object]]

    @classmethod
    def __init_subclass__(cls, **kwargs: t.Any) -> None:
        super().__init_subclass__(**kwargs)
        cls._QUOTES = _convert_quotes(cls.QUOTES)
        cls._IDENTIFIERS = _convert_quotes(cls.IDENTIFIERS)
        cls._FORMAT_STRINGS = {
            **{
                p + s: (e, TokenType.NATIONAL_STRING)
                for s, e in cls._QUOTES.items()
                for p in ("n", "N")
            },
            **_quotes_to_format(TokenType.BIT_STRING, cls.BIT_STRINGS),
            **_quotes_to_format(TokenType.BYTE_STRING, cls.BYTE_STRINGS),
            **_quotes_to_format(TokenType.HEX_STRING, cls.HEX_STRINGS),
            **_quotes_to_format(TokenType.RAW_STRING, cls.RAW_STRINGS),
            **_quotes_to_format(TokenType.HEREDOC_STRING, cls.HEREDOC_STRINGS),
            **_quotes_to_format(TokenType.UNICODE_STRING, cls.UNICODE_STRINGS),
        }
        if "BYTE_STRING_ESCAPES" not in cls.__dict__:
            cls.BYTE_STRING_ESCAPES = cls.STRING_ESCAPES.copy()
        cls._STRING_ESCAPES = set(cls.STRING_ESCAPES)
        cls._BYTE_STRING_ESCAPES = set(cls.BYTE_STRING_ESCAPES)
        cls._ESCAPE_FOLLOW_CHARS = set(cls.ESCAPE_FOLLOW_CHARS)
        cls._IDENTIFIER_ESCAPES = set(cls.IDENTIFIER_ESCAPES)
        cls._COMMENTS = {
            **{c: None for c in cls.COMMENTS if isinstance(c, str)},
            **{c[0]: c[1] for c in cls.COMMENTS if not isinstance(c, str)},
            "{#": "#}",  # Ensure Jinja comments are tokenized correctly in all dialects
        }
        if cls.HINT_START in cls.KEYWORDS:
            cls._COMMENTS[cls.HINT_START] = "*/"
        cls._KEYWORD_TRIE = new_trie(
            key.upper()
            for key in (
                *cls.KEYWORDS,
                *cls._COMMENTS,
                *cls._QUOTES,
                *cls._FORMAT_STRINGS,
            )
            if " " in key or any(single in key for single in cls.SINGLE_TOKENS)
        )


class Tokenizer(_TokenizerBase):
    SINGLE_TOKENS = {
        "(": TokenType.L_PAREN,
        ")": TokenType.R_PAREN,
        "[": TokenType.L_BRACKET,
        "]": TokenType.R_BRACKET,
        "{": TokenType.L_BRACE,
        "}": TokenType.R_BRACE,
        "&": TokenType.AMP,
        "^": TokenType.CARET,
        ":": TokenType.COLON,
        ",": TokenType.COMMA,
        ".": TokenType.DOT,
        "-": TokenType.DASH,
        "=": TokenType.EQ,
        ">": TokenType.GT,
        "<": TokenType.LT,
        "%": TokenType.MOD,
        "!": TokenType.NOT,
        "|": TokenType.PIPE,
        "+": TokenType.PLUS,
        ";": TokenType.SEMICOLON,
        "/": TokenType.SLASH,
        "\\": TokenType.BACKSLASH,
        "*": TokenType.STAR,
        "~": TokenType.TILDE,
        "?": TokenType.PLACEHOLDER,
        "@": TokenType.PARAMETER,
        "#": TokenType.HASH,
        # Used for breaking a var like x'y' but nothing else the token type doesn't matter
        "'": TokenType.UNKNOWN,
        "`": TokenType.UNKNOWN,
        '"': TokenType.UNKNOWN,
    }

    BIT_STRINGS: t.ClassVar[list[str | tuple[str, str]]] = []
    BYTE_STRINGS: t.ClassVar[list[str | tuple[str, str]]] = []
    HEX_STRINGS: t.ClassVar[list[str | tuple[str, str]]] = []
    RAW_STRINGS: t.ClassVar[list[str | tuple[str, str]]] = []
    HEREDOC_STRINGS: t.ClassVar[list[str | tuple[str, str]]] = []
    UNICODE_STRINGS: t.ClassVar[list[str | tuple[str, str]]] = []
    IDENTIFIERS: t.ClassVar[list[str | tuple[str, str]]] = ['"']
    QUOTES: t.ClassVar[list[tuple[str, str] | str]] = ["'"]
    STRING_ESCAPES: t.ClassVar[list[str]] = ["'"]
    BYTE_STRING_ESCAPES: t.ClassVar[list[str]] = []
    VAR_SINGLE_TOKENS: t.ClassVar[set[str]] = set()
    ESCAPE_FOLLOW_CHARS: t.ClassVar[list[str]] = []

    # The strings in this list can always be used as escapes, regardless of the surrounding
    # identifier delimiters. By default, the closing delimiter is assumed to also act as an
    # identifier escape, e.g. if we use double-quotes, then they also act as escapes: "x"""
    IDENTIFIER_ESCAPES: t.ClassVar[list[str]] = []

    # Whether the heredoc tags follow the same lexical rules as unquoted identifiers
    HEREDOC_TAG_IS_IDENTIFIER = False

    # Token that we'll generate as a fallback if the heredoc prefix doesn't correspond to a heredoc
    HEREDOC_STRING_ALTERNATIVE = TokenType.VAR

    # Whether string escape characters function as such when placed within raw strings
    STRING_ESCAPES_ALLOWED_IN_RAW_STRINGS = True

    NESTED_COMMENTS = True

    HINT_START = "/*+"

    TOKENS_PRECEDING_HINT = {TokenType.SELECT, TokenType.INSERT, TokenType.UPDATE, TokenType.DELETE}

    # Autofilled
    _COMMENTS: t.ClassVar[dict[str, str | None]] = {}
    _FORMAT_STRINGS: t.ClassVar[dict[str, tuple[str, TokenType]]] = {}
    _IDENTIFIERS: t.ClassVar[dict[str, str]] = {}
    _IDENTIFIER_ESCAPES: t.ClassVar[set[str]] = set()
    _QUOTES: t.ClassVar[dict[str, str]] = {}
    _STRING_ESCAPES: t.ClassVar[set[str]] = set()
    _BYTE_STRING_ESCAPES: t.ClassVar[set[str]] = set()
    _KEYWORD_TRIE: t.ClassVar[dict[str | int, object]] = {}
    _ESCAPE_FOLLOW_CHARS: t.ClassVar[set[str]] = set()

    KEYWORDS: t.ClassVar[dict[str, TokenType]] = {
        **{f"{{%{postfix}": TokenType.BLOCK_START for postfix in ("", "+", "-")},
        **{f"{prefix}%}}": TokenType.BLOCK_END for prefix in ("", "+", "-")},
        **{f"{{{{{postfix}": TokenType.BLOCK_START for postfix in ("+", "-")},
        **{f"{prefix}}}}}": TokenType.BLOCK_END for prefix in ("+", "-")},
        HINT_START: TokenType.HINT,
        "&<": TokenType.AMP_LT,
        "&>": TokenType.AMP_GT,
        "==": TokenType.EQ,
        "::": TokenType.DCOLON,
        "?::": TokenType.QDCOLON,
        "||": TokenType.DPIPE,
        "|>": TokenType.PIPE_GT,
        ">=": TokenType.GTE,
        "<=": TokenType.LTE,
        "<>": TokenType.NEQ,
        "!=": TokenType.NEQ,
        ":=": TokenType.COLON_EQ,
        "<=>": TokenType.NULLSAFE_EQ,
        "->": TokenType.ARROW,
        "->>": TokenType.DARROW,
        "=>": TokenType.FARROW,
        "#>": TokenType.HASH_ARROW,
        "#>>": TokenType.DHASH_ARROW,
        "<->": TokenType.LR_ARROW,
        "<<->>": TokenType.LLRR_ARROW,
        "&&": TokenType.DAMP,
        "??": TokenType.DQMARK,
        "~~~": TokenType.GLOB,
        "~~": TokenType.LIKE,
        "~~*": TokenType.ILIKE,
        "~*": TokenType.IRLIKE,
        "-|-": TokenType.ADJACENT,
        "ALL": TokenType.ALL,
        "AND": TokenType.AND,
        "ANTI": TokenType.ANTI,
        "ANY": TokenType.ANY,
        "ASC": TokenType.ASC,
        "AS": TokenType.ALIAS,
        "ASOF": TokenType.ASOF,
        "AUTOINCREMENT": TokenType.AUTO_INCREMENT,
        "AUTO_INCREMENT": TokenType.AUTO_INCREMENT,
        "BEGIN": TokenType.BEGIN,
        "BETWEEN": TokenType.BETWEEN,
        "CACHE": TokenType.CACHE,
        "UNCACHE": TokenType.UNCACHE,
        "CASE": TokenType.CASE,
        "CLUSTER BY": TokenType.CLUSTER_BY,
        "COLLATE": TokenType.COLLATE,
        "COLUMN": TokenType.COLUMN,
        "COMMIT": TokenType.COMMIT,
        "CONNECT BY": TokenType.CONNECT_BY,
        "CONSTRAINT": TokenType.CONSTRAINT,
        "COPY": TokenType.COPY,
        "CREATE": TokenType.CREATE,
        "CROSS": TokenType.CROSS,
        "CUBE": TokenType.CUBE,
        "CURRENT_DATE": TokenType.CURRENT_DATE,
        "CURRENT_SCHEMA": TokenType.CURRENT_SCHEMA,
        "CURRENT_TIME": TokenType.CURRENT_TIME,
        "CURRENT_TIMESTAMP": TokenType.CURRENT_TIMESTAMP,
        "CURRENT_USER": TokenType.CURRENT_USER,
        "CURRENT_CATALOG": TokenType.CURRENT_CATALOG,
        "DATABASE": TokenType.DATABASE,
        "DEFAULT": TokenType.DEFAULT,
        "DELETE": TokenType.DELETE,
        "DESC": TokenType.DESC,
        "DESCRIBE": TokenType.DESCRIBE,
        "DISTINCT": TokenType.DISTINCT,
        "DISTRIBUTE BY": TokenType.DISTRIBUTE_BY,
        "DIV": TokenType.DIV,
        "DROP": TokenType.DROP,
        "ELSE": TokenType.ELSE,
        "END": TokenType.END,
        "ENUM": TokenType.ENUM,
        "ESCAPE": TokenType.ESCAPE,
        "EXCEPT": TokenType.EXCEPT,
        "EXECUTE": TokenType.EXECUTE,
        "EXISTS": TokenType.EXISTS,
        "FALSE": TokenType.FALSE,
        "FETCH": TokenType.FETCH,
        "FILTER": TokenType.FILTER,
        "FILE": TokenType.FILE,
        "FIRST": TokenType.FIRST,
        "FULL": TokenType.FULL,
        "FUNCTION": TokenType.FUNCTION,
        "FOR": TokenType.FOR,
        "FOREIGN KEY": TokenType.FOREIGN_KEY,
        "FORMAT": TokenType.FORMAT,
        "FROM": TokenType.FROM,
        "GEOGRAPHY": TokenType.GEOGRAPHY,
        "GEOMETRY": TokenType.GEOMETRY,
        "GLOB": TokenType.GLOB,
        "GROUP BY": TokenType.GROUP_BY,
        "GROUPING SETS": TokenType.GROUPING_SETS,
        "HAVING": TokenType.HAVING,
        "ILIKE": TokenType.ILIKE,
        "IN": TokenType.IN,
        "INDEX": TokenType.INDEX,
        "INET": TokenType.INET,
        "INNER": TokenType.INNER,
        "INSERT": TokenType.INSERT,
        "INTERVAL": TokenType.INTERVAL,
        "INTERSECT": TokenType.INTERSECT,
        "INTO": TokenType.INTO,
        "IS": TokenType.IS,
        "ISNULL": TokenType.ISNULL,
        "JOIN": TokenType.JOIN,
        "KEEP": TokenType.KEEP,
        "KILL": TokenType.KILL,
        "LATERAL": TokenType.LATERAL,
        "LEFT": TokenType.LEFT,
        "LIKE": TokenType.LIKE,
        "LIMIT": TokenType.LIMIT,
        "LOAD": TokenType.LOAD,
        "LOCALTIME": TokenType.LOCALTIME,
        "LOCALTIMESTAMP": TokenType.LOCALTIMESTAMP,
        "LOCK": TokenType.LOCK,
        "MERGE": TokenType.MERGE,
        "NAMESPACE": TokenType.NAMESPACE,
        "NATURAL": TokenType.NATURAL,
        "NEXT": TokenType.NEXT,
        "NOT": TokenType.NOT,
        "NOTNULL": TokenType.NOTNULL,
        "NULL": TokenType.NULL,
        "OBJECT": TokenType.OBJECT,
        "OFFSET": TokenType.OFFSET,
        "ON": TokenType.ON,
        "OR": TokenType.OR,
        "XOR": TokenType.XOR,
        "ORDER BY": TokenType.ORDER_BY,
        "ORDINALITY": TokenType.ORDINALITY,
        "OUT": TokenType.OUT,
        "OUTER": TokenType.OUTER,
        "OVER": TokenType.OVER,
        "OVERLAPS": TokenType.OVERLAPS,
        "OVERWRITE": TokenType.OVERWRITE,
        "PARTITION": TokenType.PARTITION,
        "PARTITION BY": TokenType.PARTITION_BY,
        "PARTITIONED BY": TokenType.PARTITION_BY,
        "PARTITIONED_BY": TokenType.PARTITION_BY,
        "PERCENT": TokenType.PERCENT,
        "PIVOT": TokenType.PIVOT,
        "PRAGMA": TokenType.PRAGMA,
        "PRIMARY KEY": TokenType.PRIMARY_KEY,
        "PROCEDURE": TokenType.PROCEDURE,
        "OPERATOR": TokenType.OPERATOR,
        "QUALIFY": TokenType.QUALIFY,
        "RANGE": TokenType.RANGE,
        "RECURSIVE": TokenType.RECURSIVE,
        "REGEXP": TokenType.RLIKE,
        "RENAME": TokenType.RENAME,
        "REPLACE": TokenType.REPLACE,
        "RETURNING": TokenType.RETURNING,
        "REFERENCES": TokenType.REFERENCES,
        "RIGHT": TokenType.RIGHT,
        "RLIKE": TokenType.RLIKE,
        "ROLLBACK": TokenType.ROLLBACK,
        "ROLLUP": TokenType.ROLLUP,
        "ROW": TokenType.ROW,
        "ROWS": TokenType.ROWS,
        "SCHEMA": TokenType.SCHEMA,
        "SELECT": TokenType.SELECT,
        "SEMI": TokenType.SEMI,
        "SESSION": TokenType.SESSION,
        "SESSION_USER": TokenType.SESSION_USER,
        "SET": TokenType.SET,
        "SETTINGS": TokenType.SETTINGS,
        "SHOW": TokenType.SHOW,
        "SIMILAR TO": TokenType.SIMILAR_TO,
        "SOME": TokenType.SOME,
        "SORT BY": TokenType.SORT_BY,
        "SQL SECURITY": TokenType.SQL_SECURITY,
        "STRAIGHT_JOIN": TokenType.STRAIGHT_JOIN,
        "TABLE": TokenType.TABLE,
        "TABLESAMPLE": TokenType.TABLE_SAMPLE,
        "TEMP": TokenType.TEMPORARY,
        "TEMPORARY": TokenType.TEMPORARY,
        "THEN": TokenType.THEN,
        "TRUE": TokenType.TRUE,
        "TRUNCATE": TokenType.TRUNCATE,
        "TRIGGER": TokenType.TRIGGER,
        "UNION": TokenType.UNION,
        "UNKNOWN": TokenType.UNKNOWN,
        "UNNEST": TokenType.UNNEST,
        "UNPIVOT": TokenType.UNPIVOT,
        "UPDATE": TokenType.UPDATE,
        "USE": TokenType.USE,
        "USING": TokenType.USING,
        "UUID": TokenType.UUID,
        "VALUES": TokenType.VALUES,
        "VIEW": TokenType.VIEW,
        "VOLATILE": TokenType.VOLATILE,
        "WHEN": TokenType.WHEN,
        "WHERE": TokenType.WHERE,
        "WINDOW": TokenType.WINDOW,
        "WITH": TokenType.WITH,
        "APPLY": TokenType.APPLY,
        "ARRAY": TokenType.ARRAY,
        "BIT": TokenType.BIT,
        "BOOL": TokenType.BOOLEAN,
        "BOOLEAN": TokenType.BOOLEAN,
        "BYTE": TokenType.TINYINT,
        "MEDIUMINT": TokenType.MEDIUMINT,
        "INT1": TokenType.TINYINT,
        "TINYINT": TokenType.TINYINT,
        "INT16": TokenType.SMALLINT,
        "SHORT": TokenType.SMALLINT,
        "SMALLINT": TokenType.SMALLINT,
        "HUGEINT": TokenType.INT128,
        "UHUGEINT": TokenType.UINT128,
        "INT2": TokenType.SMALLINT,
        "INTEGER": TokenType.INT,
        "INT": TokenType.INT,
        "INT4": TokenType.INT,
        "INT32": TokenType.INT,
        "INT64": TokenType.BIGINT,
        "INT128": TokenType.INT128,
        "INT256": TokenType.INT256,
        "LONG": TokenType.BIGINT,
        "BIGINT": TokenType.BIGINT,
        "INT8": TokenType.TINYINT,
        "UINT": TokenType.UINT,
        "UINT128": TokenType.UINT128,
        "UINT256": TokenType.UINT256,
        "DEC": TokenType.DECIMAL,
        "DECIMAL": TokenType.DECIMAL,
        "DECIMAL32": TokenType.DECIMAL32,
        "DECIMAL64": TokenType.DECIMAL64,
        "DECIMAL128": TokenType.DECIMAL128,
        "DECIMAL256": TokenType.DECIMAL256,
        "DECFLOAT": TokenType.DECFLOAT,
        "BIGDECIMAL": TokenType.BIGDECIMAL,
        "BIGNUMERIC": TokenType.BIGDECIMAL,
        "BIGNUM": TokenType.BIGNUM,
        "LIST": TokenType.LIST,
        "MAP": TokenType.MAP,
        "NULLABLE": TokenType.NULLABLE,
        "NUMBER": TokenType.DECIMAL,
        "NUMERIC": TokenType.DECIMAL,
        "FIXED": TokenType.DECIMAL,
        "REAL": TokenType.FLOAT,
        "FLOAT": TokenType.FLOAT,
        "FLOAT4": TokenType.FLOAT,
        "FLOAT8": TokenType.DOUBLE,
        "DOUBLE": TokenType.DOUBLE,
        "DOUBLE PRECISION": TokenType.DOUBLE,
        "JSON": TokenType.JSON,
        "JSONB": TokenType.JSONB,
        "CHAR": TokenType.CHAR,
        "CHARACTER": TokenType.CHAR,
        "CHAR VARYING": TokenType.VARCHAR,
        "CHARACTER VARYING": TokenType.VARCHAR,
        "NCHAR": TokenType.NCHAR,
        "VARCHAR": TokenType.VARCHAR,
        "VARCHAR2": TokenType.VARCHAR,
        "NVARCHAR": TokenType.NVARCHAR,
        "NVARCHAR2": TokenType.NVARCHAR,
        "BPCHAR": TokenType.BPCHAR,
        "STR": TokenType.TEXT,
        "STRING": TokenType.TEXT,
        "TEXT": TokenType.TEXT,
        "LONGTEXT": TokenType.LONGTEXT,
        "MEDIUMTEXT": TokenType.MEDIUMTEXT,
        "TINYTEXT": TokenType.TINYTEXT,
        "CLOB": TokenType.TEXT,
        "LONGVARCHAR": TokenType.TEXT,
        "BINARY": TokenType.BINARY,
        "BLOB": TokenType.VARBINARY,
        "LONGBLOB": TokenType.LONGBLOB,
        "MEDIUMBLOB": TokenType.MEDIUMBLOB,
        "TINYBLOB": TokenType.TINYBLOB,
        "BYTEA": TokenType.VARBINARY,
        "VARBINARY": TokenType.VARBINARY,
        "TIME": TokenType.TIME,
        "TIMETZ": TokenType.TIMETZ,
        "TIME_NS": TokenType.TIME_NS,
        "TIMESTAMP": TokenType.TIMESTAMP,
        "TIMESTAMPTZ": TokenType.TIMESTAMPTZ,
        "TIMESTAMPLTZ": TokenType.TIMESTAMPLTZ,
        "TIMESTAMP_LTZ": TokenType.TIMESTAMPLTZ,
        "TIMESTAMPNTZ": TokenType.TIMESTAMPNTZ,
        "TIMESTAMP_NTZ": TokenType.TIMESTAMPNTZ,
        "DATE": TokenType.DATE,
        "DATETIME": TokenType.DATETIME,
        "INT4RANGE": TokenType.INT4RANGE,
        "INT4MULTIRANGE": TokenType.INT4MULTIRANGE,
        "INT8RANGE": TokenType.INT8RANGE,
        "INT8MULTIRANGE": TokenType.INT8MULTIRANGE,
        "NUMRANGE": TokenType.NUMRANGE,
        "NUMMULTIRANGE": TokenType.NUMMULTIRANGE,
        "TSRANGE": TokenType.TSRANGE,
        "TSMULTIRANGE": TokenType.TSMULTIRANGE,
        "TSTZRANGE": TokenType.TSTZRANGE,
        "TSTZMULTIRANGE": TokenType.TSTZMULTIRANGE,
        "DATERANGE": TokenType.DATERANGE,
        "DATEMULTIRANGE": TokenType.DATEMULTIRANGE,
        "UNIQUE": TokenType.UNIQUE,
        "VECTOR": TokenType.VECTOR,
        "STRUCT": TokenType.STRUCT,
        "SEQUENCE": TokenType.SEQUENCE,
        "VARIANT": TokenType.VARIANT,
        "ALTER": TokenType.ALTER,
        "ANALYZE": TokenType.ANALYZE,
        "CALL": TokenType.COMMAND,
        "COMMENT": TokenType.COMMENT,
        "EXPLAIN": TokenType.COMMAND,
        "GRANT": TokenType.GRANT,
        "REVOKE": TokenType.REVOKE,
        "OPTIMIZE": TokenType.COMMAND,
        "PREPARE": TokenType.COMMAND,
        "VACUUM": TokenType.COMMAND,
        "USER-DEFINED": TokenType.USERDEFINED,
    }

    COMMANDS = {
        TokenType.COMMAND,
        TokenType.EXECUTE,
        TokenType.FETCH,
        TokenType.SHOW,
        TokenType.RENAME,
    }

    COMMAND_PREFIX_TOKENS = {TokenType.SEMICOLON, TokenType.BEGIN}

    # Handle numeric literals like in hive (3L = BIGINT)
    NUMERIC_LITERALS: t.ClassVar[dict[str, str]] = {}

    # In tokenizers like JSONPath, dots are always key separators, never decimal points
    NUMBERS_CAN_HAVE_DECIMALS: t.ClassVar[bool] = True

    COMMENTS = ["--", ("/*", "*/")]

    __slots__ = (
        "dialect",
        "_core",
    )

    def __init__(self, dialect: DialectType = None) -> None:
        from sqlglot.dialects.dialect import Dialect

        self.dialect = Dialect.get_or_raise(dialect)
        self._core = self._init_core()

    def _init_core(self) -> TokenizerCore:
        return TokenizerCore(
            single_tokens=self.SINGLE_TOKENS,
            keywords=self.KEYWORDS,
            quotes=self._QUOTES,
            format_strings=self._FORMAT_STRINGS,
            identifiers=self._IDENTIFIERS,
            comments=self._COMMENTS,
            string_escapes=self._STRING_ESCAPES,
            byte_string_escapes=self._BYTE_STRING_ESCAPES,
            identifier_escapes=self._IDENTIFIER_ESCAPES,
            escape_follow_chars=self._ESCAPE_FOLLOW_CHARS,
            commands=self.COMMANDS,
            command_prefix_tokens=self.COMMAND_PREFIX_TOKENS,
            nested_comments=self.NESTED_COMMENTS,
            hint_start=self.HINT_START,
            tokens_preceding_hint=self.TOKENS_PRECEDING_HINT,
            has_bit_strings=bool(self.BIT_STRINGS),
            has_hex_strings=bool(self.HEX_STRINGS),
            numeric_literals=self.NUMERIC_LITERALS,
            var_single_tokens=self.VAR_SINGLE_TOKENS,
            string_escapes_allowed_in_raw_strings=self.STRING_ESCAPES_ALLOWED_IN_RAW_STRINGS,
            heredoc_tag_is_identifier=self.HEREDOC_TAG_IS_IDENTIFIER,
            heredoc_string_alternative=self.HEREDOC_STRING_ALTERNATIVE,
            keyword_trie=self._KEYWORD_TRIE,
            numbers_can_be_underscore_separated=self.dialect.NUMBERS_CAN_BE_UNDERSCORE_SEPARATED,
            numbers_can_have_decimals=self.NUMBERS_CAN_HAVE_DECIMALS,
            identifiers_can_start_with_digit=self.dialect.IDENTIFIERS_CAN_START_WITH_DIGIT,
            unescaped_sequences=self.dialect.UNESCAPED_SEQUENCES,
        )

    def tokenize(self, sql: str) -> list[Token]:
        """Returns a list of tokens corresponding to the SQL string `sql`."""
        return self._core.tokenize(sql)  # type: ignore

    @property
    def sql(self) -> str:
        """The SQL string being tokenized."""
        return self._core.sql

    @property
    def size(self) -> int:
        """Length of the SQL string."""
        return self._core.size

    @property
    def tokens(self) -> list[Token]:
        """The list of tokens produced by tokenization."""
        return self._core.tokens



import re
import unittest
import sys

from sqlglot.dialects.dialect import Dialect, Dialects
from sqlglot.generators.hive import CANONICAL_TIME_FORMAT, LAX_TO_NON_PADDED_FORMATS
from sqlglot.time import format_time, subsecond_precision

KNOWN_CANONICAL_ATOM = re.compile(r"%(?:[mdHIMS]strict|[-:].|[\w%])")


class TestTime(unittest.TestCase):
    def test_format_time(self):
        self.assertEqual(format_time("", {}), None)
        self.assertEqual(format_time(" ", {}), " ")
        mapping = {"a": "b", "aa": "c"}
        self.assertEqual(format_time("a", mapping), "b")
        self.assertEqual(format_time("aa", mapping), "c")
        self.assertEqual(format_time("aaada", mapping), "cbdb")
        self.assertEqual(format_time("da", mapping), "db")

    def test_canonical_time_format_vocabulary(self):
        # _lenient_parse_format assumes CANONICAL_TIME_FORMAT tokenizes every canonical time
        # format atomically; fail loudly when a dialect introduces an atom it can't recognize,
        # instead of silently mis-rewriting formats generated for strict dialects
        for dialect in Dialects:
            if not dialect.value:
                continue

            d = Dialect.get_or_raise(dialect.value)
            for value in (*d.TIME_MAPPING.values(), *d.FORMAT_MAPPING.values()):
                for part in CANONICAL_TIME_FORMAT.split(value):
                    self.assertNotIn("%", part, f"{dialect.value}: unconsumed '%' in {value!r}")
                for atom in CANONICAL_TIME_FORMAT.findall(value):
                    self.assertTrue(
                        KNOWN_CANONICAL_ATOM.fullmatch(atom),
                        f"{dialect.value}: unexpected canonical atom {atom!r} in {value!r} - "
                        "update CANONICAL_TIME_FORMAT in sqlglot/generators/hive.py",
                    )
                self.assertIsNone(
                    re.search(r"%[mdHIMS](?!strict)\w", value),
                    f"{dialect.value}: a lax %m/%d/%H/%I/%M/%S directly followed by a word "
                    f"character in {value!r} would be mis-tokenized by CANONICAL_TIME_FORMAT",
                )

            if d.TIME_MAPPING.get("MM") == "%mstrict":
                # _lenient_parse_format rewrites the format before the inverse trie walk, so
                # strict dialects must not rely on composite inverse keys containing a lax specifier
                lax = set(LAX_TO_NON_PADDED_FORMATS)
                for key in d.INVERSE_TIME_MAPPING:
                    atoms = CANONICAL_TIME_FORMAT.findall(key)
                    if len(atoms) > 1:
                        self.assertFalse(
                            lax & set(atoms),
                            f"{dialect.value}: composite inverse key {key!r} contains a lax specifier",
                        )

    def test_subsecond_precision(self):
        self.assertEqual(6, subsecond_precision("2023-01-01 12:13:14.123456+00:00"))
        self.assertEqual(6, subsecond_precision("2023-01-01 12:13:14.000123+00:00"))
        self.assertEqual(6, subsecond_precision("2023-01-01 12:13:14.000001+00:00"))
        self.assertEqual(3, subsecond_precision("2023-01-01 12:13:14.123+00:00"))
        self.assertEqual(0, subsecond_precision("2023-01-01 12:13:14+00:00"))
        self.assertEqual(0, subsecond_precision("2023-01-01 12:13:14"))
        self.assertEqual(0, subsecond_precision("garbage"))

    @unittest.skipUnless(
        sys.version_info >= (3, 11),
        "Python 3.11 relaxed datetime.fromisoformat() parsing with regards to microseconds",
    )
    def test_subsecond_precision_python311(self):
        # ref: https://docs.python.org/3/whatsnew/3.11.html#datetime
        self.assertEqual(6, subsecond_precision("2023-01-01 12:13:14.123456789+00:00"))
        self.assertEqual(6, subsecond_precision("2023-01-01 12:13:14.12345+00:00"))
        self.assertEqual(6, subsecond_precision("2023-01-01 12:13:14.1234+00:00"))
        self.assertEqual(3, subsecond_precision("2023-01-01 12:13:14.12+00:00"))
        self.assertEqual(3, subsecond_precision("2023-01-01 12:13:14.1+00:00"))



from __future__ import annotations

from sqlglot import exp, parser
from sqlglot.tokens import TokenType


class DAXParser(parser.Parser):
    CONJUNCTION = {
        **parser.Parser.CONJUNCTION,
        TokenType.DAMP: exp.And,
    }

    DISJUNCTION = {
        **parser.Parser.DISJUNCTION,
        TokenType.DPIPE: exp.Or,
    }

    def _parse_statement(self) -> exp.Expr | None:
        if not self._match_text_seq("EVALUATE"):
            return None

        query = self._parse_table_expression()

        if query and self._match(TokenType.ORDER_BY):
            order = exp.Order(expressions=self._parse_csv(self._parse_ordered))
            query = query.order_by(self.expression(order), copy=False)

        return query

    def _parse_table_expression(self) -> exp.Query | None:
        if self._match_pair(TokenType.FILTER, TokenType.L_PAREN):
            query = self._parse_table_expression()
            self._match(TokenType.COMMA)
            where = self._parse_disjunction()
            self._match_r_paren()
            return query.where(where, copy=False) if query else None

        table = self._parse_table_parts()
        return exp.select("*").from_(table, copy=False) if table else None

    def _parse_column(self) -> exp.Expr | None:
        this = super()._parse_column()

        # Combine adjacent identifiers into a column reference: Table[Column]
        if isinstance(this, exp.Column) and (column := self._parse_identifier()):
            this = self.expression(exp.Column(this=column, table=this.this))

        return this



from __future__ import annotations

from sqlglot import exp
from sqlglot.helper import seq_get
from sqlglot.generators.postgres import PostgresGenerator

from sqlglot.transforms import (
    remove_unique_constraints,
    ctas_with_tmp_tables_to_create_tmp_view,
    preprocess,
)


class MaterializeGenerator(PostgresGenerator):
    SUPPORTS_CREATE_TABLE_LIKE = False
    SUPPORTS_BETWEEN_FLAGS = False

    TRANSFORMS = {
        **{k: v for k, v in PostgresGenerator.TRANSFORMS.items() if k != exp.ToMap},
        exp.AutoIncrementColumnConstraint: lambda self, e: "",
        exp.Create: preprocess(
            [
                remove_unique_constraints,
                ctas_with_tmp_tables_to_create_tmp_view,
            ]
        ),
        exp.GeneratedAsIdentityColumnConstraint: lambda self, e: "",
        exp.OnConflict: lambda self, e: "",
        exp.PrimaryKeyColumnConstraint: lambda self, e: "",
    }

    def propertyeq_sql(self, expression: exp.PropertyEQ) -> str:
        return self.binary(expression, "=>")

    def datatype_sql(self, expression: exp.DataType) -> str:
        if expression.is_type(exp.DType.LIST):
            if expression.expressions:
                return f"{self.expressions(expression, flat=True)} LIST"
            return "LIST"

        if expression.is_type(exp.DType.MAP) and len(expression.expressions) == 2:
            key, value = expression.expressions
            return f"MAP[{self.sql(key)} => {self.sql(value)}]"

        return super().datatype_sql(expression)

    def list_sql(self, expression: exp.List) -> str:
        if isinstance(seq_get(expression.expressions, 0), exp.Select):
            return self.func("LIST", seq_get(expression.expressions, 0))

        return f"{self.normalize_func('LIST')}[{self.expressions(expression, flat=True)}]"

    def tomap_sql(self, expression: exp.ToMap) -> str:
        if isinstance(expression.this, exp.Select):
            return self.func("MAP", expression.this)
        return f"{self.normalize_func('MAP')}[{self.expressions(expression.this)}]"



"""Refleak detector for sqlglot[c].

Runs the unit-test suite N times against the compiled build, snapshotting
``gc.get_objects()`` type counts between runs and flagging types whose count
grows monotonically with positive slope. Also reports ``sys.gettotalrefcount``
drift (debug Python only) and RSS drift as coarse backstops.

Why three signals, in order of strength:

1. **Per-type object-count growth**. Works on every Python build. Catches
   leaks of gc-tracked objects — the class of leak mypyc refcount bugs
   typically produce, because mypyc-generated classes are proper PyObject
   types registered with the gc.
2. **``sys.gettotalrefcount`` drift**. Debug Python only (``--with-pydebug``).
   Catches leaks of non-gc-tracked objects (small ints, short strings,
   interned tuples) that signal #1 misses.
3. **RSS drift**. OS-level backstop for raw-malloc leaks that never show up
   in CPython bookkeeping at all (e.g. unreleased arenas, third-party C
   libraries).

The workload is the real unit test suite — same as ``make unitc``. Tests are
loaded once and re-run in-process each iteration; running them out-of-process
would reset the heap and hide the growth we're trying to see.

Usage:
    python -m tests.leakcheck                  # discover tests/ pattern test_*.py
    python -m tests.leakcheck --top tests/dialects
    python -m tests.leakcheck --pattern test_snowflake.py
    python -m tests.leakcheck --warmup 2 --samples 5 --verbose
"""

from __future__ import annotations

import argparse
import gc
import os
import resource
import sys
import unittest
from collections import Counter


# Built-in bookkeeping types whose counts fluctuate for reasons unrelated to
# the workload (stack frames come and go, tracebacks get created lazily,
# iterators are ephemeral). Ignoring them keeps the "Growing types" report
# focused on domain-meaningful types like Expression subclasses.
#
# unittest-machinery entries here: each TextTestRunner.run() call allocates a
# fresh result/runner/stream wrapper, and the suite's subtest tracking tracks
# new _SubTest/_OrderedChainMap per test. These grow linearly with iteration
# count purely because we re-run the suite, not because sqlglot leaks.
_NOISY = frozenset(
    {
        "frame",
        "traceback",
        "cell",
        "function",
        "method",
        "builtin_function_or_method",
        "method_descriptor",
        "wrapper_descriptor",
        "member_descriptor",
        "getset_descriptor",
        "classmethod_descriptor",
        "list_iterator",
        "tuple_iterator",
        "dict_keyiterator",
        "dict_valueiterator",
        "dict_itemiterator",
        "set_iterator",
        # unittest internals re-allocated per run
        "TestResult",
        "TextTestResult",
        "TextTestRunner",
        "_WritelnDecorator",
        "_SubTest",
        "_OrderedChainMap",
        "_GeneratorContextManager",
        "_Outcome",
        "generator",
        "ReferenceType",
    }
)

# Structural containers are load-bearing but legitimately grow from lazy
# caches (dialect registries, normalized-name caches, etc. keep filling).
# Treat them with a higher delta threshold so first-run cache warmup doesn't
# flip the verdict; a real leak still wins because its slope is linear and
# unbounded.
_STRUCTURAL = frozenset({"dict", "list", "tuple", "set", "frozenset"})


# Opened once at import rather than per-iteration to avoid FD churn and to
# keep it out of the per-sample object count.
_NULL_STREAM = open(os.devnull, "w")


def _maxrss_kb() -> int:
    """Return current resident-set size in KB, normalized across OSes.

    ``resource.ru_maxrss`` is specified in KB on Linux but in bytes on macOS
    (and BSD). Normalizing here lets the RSS-drift threshold be a single
    value regardless of platform.
    """
    raw = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss
    return raw // 1024 if sys.platform == "darwin" else raw


def _snapshot():
    """Snapshot the live Python heap: per-type counts, totalrefcount, RSS.

    Returns ``(counts, totalrefcount, maxrss_kb)`` where ``counts`` is
    ``{type_name: count}`` over every gc-tracked object alive right now.

    **Why two ``gc.collect`` calls.** A single collection can leave work for
    a second pass in two situations:

    * Weakref callbacks fire during collection and can drop the last strong
      reference to an object; that object is then only freed on the next
      pass.
    * Objects with finalizers (``__del__``) are moved to ``gc.garbage``-like
      handling per PEP 442 and may only be reclaimed in a subsequent cycle.

    Without the second pass the snapshot would include objects that are
    logically dead but not yet freed — iteration-to-iteration that "live"
    set fluctuates, producing non-monotonic noise that masks the linear
    leak signal we're trying to isolate. Two collects makes the heap state
    stable enough that the growth we measure reflects real retention.

    **Why ``sys.gettotalrefcount`` is probed via ``getattr``.** It only
    exists on debug Python builds; on stock CPython the symbol is absent
    and ``getattr(..., None)()`` returns ``None`` so the caller can treat
    the signal as "unavailable" rather than raise.
    """
    gc.collect()
    gc.collect()
    counts: Counter[str] = Counter()
    for obj in gc.get_objects():
        counts[type(obj).__name__] += 1
    rc = getattr(sys, "gettotalrefcount", lambda: None)()
    return dict(counts), rc, _maxrss_kb()


def _slope(values):
    """Least-squares slope of ``values`` against their index.

    Used as the primary leak-signal metric rather than ``values[-1] -
    values[0]`` because:

    * A per-iteration-constant leak produces a straight line (``+k, +k,
      +k, ...``); least-squares slope recovers ``k`` exactly.
    * Averaging across all samples is robust to a single noisy outlier
      that a first-vs-last delta would overweight.
    * A plateauing cache (``+k, +k, 0, 0, 0, ...``) produces a lower
      slope than a sustained leak of the same magnitude, so the
      threshold-based check naturally prefers the latter.
    """
    n = len(values)
    mean_x = (n - 1) / 2
    mean_y = sum(values) / n
    num = sum((i - mean_x) * (v - mean_y) for i, v in enumerate(values))
    den = sum((i - mean_x) ** 2 for i in range(n))
    return num / den if den else 0.0


def check_leaks(
    workload,
    *,
    warmup=1,
    samples=3,
    slope_min=0.5,
    # Full-suite workloads have a small baseline growth from test fixtures
    # that register module-level state (e.g. tests that declare new Dialect
    # subclasses). We observe ~3 such allocations per iteration in a clean
    # build, so delta_min=10 puts real leaks (always far louder) above the
    # floor while absorbing fixture noise.
    delta_min=10,
    # Structural containers (dict/list/tuple/set/frozenset) get a higher bar
    # because the full unit-test suite legitimately allocates thousands of
    # them per run (TestResult fields, collected-tests lists, etc.). A real
    # leak still wins — the post1 property-getter bug produced dict delta
    # ~750k per 3 samples, two orders of magnitude above this threshold.
    delta_min_structural=10000,
    # ``ru_maxrss`` is a peak high-water mark, not current RSS — once the
    # process touches a page, the counter never goes back down. Running the
    # full unit suite repeatedly causes significant variance from arena
    # fragmentation and test-specific working sets (we've observed 20 MB on
    # one run and 470 MB on the next). Treat RSS as a coarse backstop for
    # truly pathological growth rather than a precise gate.
    rss_threshold_kb=1_000_000,
    refcount_threshold=2000,
):
    """Run ``workload`` repeatedly and return growth findings.

    Execution model:

    * **Warmup phase** — ``warmup`` runs without sampling, letting lazy
      module-level state converge (dialect class registries, tokenizer
      trie construction, normalized-name caches, etc.). Without warmup
      the first few "samples" would conflate first-use cache population
      with real leak growth.
    * **Sampling phase** — ``samples`` more runs, snapshotting the heap
      both before the first run and after each subsequent one. The
      resulting per-type series is what the detection logic operates on.

    A type is flagged if **all three** hold:

    1. Its count is strictly non-decreasing across samples. Any
       oscillation rules it out as noise or churn — caches sometimes
       shrink briefly under GC, real leaks never do.
    2. Its least-squares slope is at least ``slope_min``. A constant
       slope is the fingerprint of a per-call leak of a fixed number
       of objects.
    3. Its total delta is at least ``delta_min`` (or
       ``delta_min_structural`` for dict/list/tuple/set/frozenset, which
       grow legitimately during cache warmup and need a stricter bar).

    Also computed:

    * ``rc_delta`` — ``sys.gettotalrefcount`` drift from first to last
      snapshot. ``None`` on release Python builds.
    * ``rss_delta`` — process RSS growth in KB.

    The overall verdict is ``leaked = bool(reasons)`` where ``reasons``
    aggregates any of: growing types found, refcount drift above
    ``refcount_threshold``, RSS drift above ``rss_threshold_kb``.

    Returns ``(growing, rc_delta, rss_delta_kb, reasons)``. ``growing``
    is ``[(type_name, slope, delta, values), ...]`` sorted by slope
    descending; ``reasons`` is a list of short strings (empty ⇒ clean).
    """
    for _ in range(warmup):
        workload()

    snaps = [_snapshot()]
    for _ in range(samples):
        workload()
        snaps.append(_snapshot())

    # Union of every type that appeared in any snapshot — a type can first
    # appear mid-run (lazy import), so we can't just take the first snapshot.
    names: set[str] = set()
    for counts, _, _ in snaps:
        names.update(counts)

    growing = []
    for tname in sorted(names - _NOISY):
        vals = [c.get(tname, 0) for c, _, _ in snaps]
        if not all(vals[i + 1] >= vals[i] for i in range(len(vals) - 1)):
            continue
        slope = _slope(vals)
        delta = vals[-1] - vals[0]
        thresh = delta_min_structural if tname in _STRUCTURAL else delta_min
        if slope >= slope_min and delta >= thresh:
            growing.append((tname, slope, delta, vals))
    # Steepest slopes first so truncated reports keep the most-interesting rows.
    growing.sort(key=lambda g: -g[1])

    rc0, rc1 = snaps[0][1], snaps[-1][1]
    rc_delta = rc1 - rc0 if rc0 is not None else None
    rss_delta = snaps[-1][2] - snaps[0][2]

    reasons = []
    if growing:
        reasons.append(f"{len(growing)} type(s) growing")
    if rc_delta is not None and rc_delta >= refcount_threshold:
        reasons.append(f"totalrefcount +{rc_delta}")
    if rss_delta >= rss_threshold_kb:
        reasons.append(f"RSS +{rss_delta} KB")

    return growing, rc_delta, rss_delta, reasons


def _print_report(name, growing, rc_delta, rss_delta, reasons, *, verbose=False):
    """Format the tuple returned by ``check_leaks`` to stdout.

    Default truncates at 15 growing types; ``verbose=True`` prints all of
    them. Each row shows the full per-iteration count trail so the reader
    can eyeball monotonicity without re-running.
    """
    status = "LEAK" if reasons else " OK "
    header = f"[{status}] {name}"
    if reasons:
        header += "  — " + "; ".join(reasons)
    print(header)
    if growing:
        print("  Growing object types:")
        shown = growing if verbose else growing[:15]
        for tname, slope, delta, vals in shown:
            trail = " ".join(str(v) for v in vals)
            print(f"    {tname:<36} slope={slope:+6.2f}/iter  delta={delta:+d}   [{trail}]")
        if not verbose and len(growing) > len(shown):
            print(f"    ... and {len(growing) - len(shown)} more")
    if rc_delta is not None:
        print(f"  sys.gettotalrefcount delta: {rc_delta:+d}")
    print(f"  maxrss delta: {rss_delta:+d} KB")


def _make_workload(top, pattern):
    """Return a zero-arg callable that re-runs the unit test suite.

    The suite is discovered **once** so the ``type`` objects for each
    TestCase subclass are created exactly one time; a fresh discover per
    iteration would inflate the ``type`` count and produce false-positive
    growth on that row.

    ``unittest.BaseTestSuite._cleanup`` defaults to ``True``: after each
    test runs, the suite sets its slot to ``None`` to drop the reference
    (a memory optimisation for normal single-shot runs). That would leave
    the second iteration running zero tests. We disable it class-wide so
    the same suite object can be replayed N times.
    """
    unittest.BaseTestSuite._cleanup = False
    # Explicit top_level_dir prevents Python 3.14's discover() from prepending
    # the tests/ directory to sys.path ahead of the project root, which would
    # let tests/sqlglot/ shadow the real sqlglot package in multiprocessing workers.
    top_level = os.path.normpath(os.path.join(os.path.dirname(__file__), ".."))
    suite = unittest.TestLoader().discover(top, pattern=pattern, top_level_dir=top_level)

    def run():
        unittest.TextTestRunner(stream=_NULL_STREAM, verbosity=0).run(suite)

    return run


def main():
    """CLI entry point. Exits 1 if a leak was flagged, else 0."""
    p = argparse.ArgumentParser(description=__doc__.splitlines()[0])
    p.add_argument("--top", default="tests", help="Test discovery root (default: tests)")
    p.add_argument("--pattern", default="test_*.py", help="Test file glob (default: test_*.py)")
    p.add_argument("--warmup", type=int, default=1)
    p.add_argument("--samples", type=int, default=3)
    p.add_argument("--verbose", action="store_true")
    args = p.parse_args()

    # Integration tests hit external DBs and are flaky/slow — unfit for a
    # harness that runs the suite multiple times. ``setdefault`` lets the
    # caller opt in explicitly by pre-exporting ``SKIP_INTEGRATION=0``.
    os.environ.setdefault("SKIP_INTEGRATION", "1")

    # Detect whether we're running against the compiled extension. The .so
    # is resolved by the normal import machinery (it wins over the co-located
    # .py), so probing any mypyc-compiled module's __file__ is sufficient.
    import sqlglot.expressions.core as _ec

    compiled = _ec.__file__.endswith(".so")
    print(f"build: {'sqlglotc (compiled)' if compiled else 'pure python'}")
    print(f"debug Python: {'yes' if hasattr(sys, 'gettotalrefcount') else 'no'}")
    print(f"workload: unittest discover {args.top} pattern {args.pattern}")
    print(f"warmup={args.warmup} samples={args.samples}")
    print()

    workload = _make_workload(args.top, args.pattern)
    growing, rc_delta, rss_delta, reasons = check_leaks(
        workload, warmup=args.warmup, samples=args.samples
    )
    _print_report(
        f"{args.top} ({args.pattern})",
        growing,
        rc_delta,
        rss_delta,
        reasons,
        verbose=args.verbose,
    )
    return 1 if reasons else 0


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

import typing as t

from sqlglot import exp, parser
from sqlglot.dialects.dialect import build_formatted_time, build_timetostr_or_tochar, build_trunc
from sqlglot.helper import ensure_list, seq_get
from sqlglot.parser import OPTIONS_TYPE, build_coalesce
from sqlglot.tokens import TokenType

if t.TYPE_CHECKING:
    from sqlglot._typing import E
    from sqlglot.dialects.dialect import Dialect


def _build_to_timestamp(args: list, dialect: Dialect) -> exp.StrToTime | exp.Anonymous:
    if len(args) == 1:
        return exp.Anonymous(this="TO_TIMESTAMP", expressions=args)

    return build_formatted_time(exp.StrToTime)(args, dialect)


class OracleParser(parser.Parser):
    WINDOW_BEFORE_PAREN_TOKENS = {TokenType.OVER, TokenType.KEEP}
    VALUES_FOLLOWED_BY_PAREN = False

    FUNCTIONS = {
        **{k: v for k, v in parser.Parser.FUNCTIONS.items() if k != "TO_BOOLEAN"},
        "CONVERT": exp.ConvertToCharset.from_arg_list,
        "L2_DISTANCE": exp.EuclideanDistance.from_arg_list,
        "NVL": lambda args: build_coalesce(args, is_nvl=True),
        "SQUARE": lambda args: exp.Pow(this=seq_get(args, 0), expression=exp.Literal.number(2)),
        "TO_CHAR": build_timetostr_or_tochar,
        "TO_TIMESTAMP": _build_to_timestamp,
        "TO_DATE": build_formatted_time(exp.StrToDate),
        "TRUNC": lambda args, dialect: build_trunc(
            args, dialect, date_trunc_unabbreviate=False, default_date_trunc_unit="DD"
        ),
    }

    NO_PAREN_FUNCTION_PARSERS = {
        **parser.Parser.NO_PAREN_FUNCTION_PARSERS,
        "NEXT": lambda self: self._parse_next_value_for(),
        "PRIOR": lambda self: self.expression(exp.Prior(this=self._parse_bitwise())),
        "SYSDATE": lambda self: self.expression(exp.CurrentTimestamp(sysdate=True)),
        "DBMS_RANDOM": lambda self: self._parse_dbms_random(),
    }

    NO_PAREN_FUNCTIONS = {
        **parser.Parser.NO_PAREN_FUNCTIONS,
        TokenType.LOCALTIMESTAMP: exp.Localtimestamp,
        TokenType.SYSTIMESTAMP: exp.Systimestamp,
    }

    FUNCTION_PARSERS = {
        **{k: v for k, v in parser.Parser.FUNCTION_PARSERS.items() if k != "CONVERT"},
        "JSON_ARRAY": lambda self: self._parse_oracle_json_array(),
        "JSON_ARRAYAGG": lambda self: self._parse_oracle_json_arrayagg(),
        "JSON_EXISTS": lambda self: self._parse_json_exists(),
        "LISTAGG": lambda self: self._parse_string_agg(),
        "TO_NUMBER": lambda self: self._parse_to_number(),
    }

    PROPERTY_PARSERS = {
        **parser.Parser.PROPERTY_PARSERS,
        "GLOBAL": lambda self: (
            self._match_text_seq("TEMPORARY")
            and self.expression(exp.TemporaryProperty(this="GLOBAL"))
        ),
        "PRIVATE": lambda self: (
            self._match_text_seq("TEMPORARY")
            and self.expression(exp.TemporaryProperty(this="PRIVATE"))
        ),
        "FORCE": lambda self: self.expression(exp.ForceProperty()),
    }

    QUERY_MODIFIER_PARSERS = {
        **parser.Parser.QUERY_MODIFIER_PARSERS,
        TokenType.ORDER_SIBLINGS_BY: lambda self: ("order", self._parse_order()),
        TokenType.WITH: lambda self: (
            "options",
            ensure_list(self._parse_query_restrictions()),
        ),
    }

    TYPE_LITERAL_PARSERS = {
        exp.DType.DATE: lambda self, this, _: self.expression(exp.DateStrToDate(this=this)),
        # https://docs.oracle.com/en/database/oracle/oracle-database/19/refrn/NLS_TIMESTAMP_FORMAT.html
        exp.DType.TIMESTAMP: lambda self, this, _: _build_to_timestamp(
            [this, '"%Y-%m-%d %H:%M:%S.%f"'], self.dialect
        ),
    }

    # SELECT UNIQUE .. is old-style Oracle syntax for SELECT DISTINCT ..
    # Reference: https://stackoverflow.com/a/336455
    DISTINCT_TOKENS = {TokenType.DISTINCT, TokenType.UNIQUE}

    QUERY_RESTRICTIONS: t.ClassVar[OPTIONS_TYPE] = {
        "WITH": (
            ("READ", "ONLY"),
            ("CHECK", "OPTION"),
        ),
    }

    def _parse_to_number(self) -> exp.ToNumber:
        # https://docs.oracle.com/en/database/oracle/oracle-database/19/sqlrf/TO_NUMBER.html
        this = self._parse_bitwise()

        default = None
        if self._match(TokenType.DEFAULT):
            default = self._parse_bitwise()
            self._match_text_seq("ON", "CONVERSION", "ERROR")

        fmt = None
        nlsparam = None
        if self._match(TokenType.COMMA):
            fmt = self._parse_bitwise()
            if self._match(TokenType.COMMA):
                nlsparam = self._parse_bitwise()

        return self.expression(
            exp.ToNumber(this=this, format=fmt, nlsparam=nlsparam, default=default)
        )

    def _parse_dbms_random(self) -> exp.Expr | None:
        if self._match_text_seq(".", "VALUE"):
            lower, upper = None, None
            if self._match(TokenType.L_PAREN, advance=False):
                lower_upper = self._parse_wrapped_csv(self._parse_bitwise)
                if len(lower_upper) == 2:
                    lower, upper = lower_upper

            return exp.Rand(lower=lower, upper=upper)

        self._retreat(self._index - 1)
        return None

    def _parse_oracle_json_array(self) -> exp.JSONArray:
        return self._parse_json_array(
            exp.JSONArray,
            expressions=self._parse_csv(lambda: self._parse_format_json(self._parse_bitwise())),
        )

    def _parse_oracle_json_arrayagg(self) -> exp.JSONArrayAgg:
        return self._parse_json_array(
            exp.JSONArrayAgg,
            this=self._parse_format_json(self._parse_bitwise()),
            order=self._parse_order(),
        )

    def _parse_json_array(self, expr_type: type[E], **kwargs) -> E:
        return self.expression(
            expr_type(
                null_handling=self._parse_on_handling("NULL", "NULL", "ABSENT"),
                return_type=self._match_text_seq("RETURNING") and self._parse_type(),
                strict=self._match_text_seq("STRICT"),
                **kwargs,
            )
        )

    def _parse_hint_function_call(self) -> exp.Expr | None:
        if not self._curr or not self._next or self._next.token_type != TokenType.L_PAREN:
            return None

        name = self._curr.text

        self._advance(2)
        args = self._parse_hint_args()
        this = self.expression(exp.Anonymous(this=name, expressions=args))
        self._match_r_paren(this)
        return this

    def _parse_hint_args(self):
        args = []
        result = self._parse_var()

        while result:
            args.append(result)
            result = self._parse_var()

        return args

    def _parse_query_restrictions(self) -> exp.Expr | None:
        kind = self._parse_var_from_options(self.QUERY_RESTRICTIONS, raise_unmatched=False)

        if not kind:
            return None

        return self.expression(
            exp.QueryOption(
                this=kind, expression=self._match(TokenType.CONSTRAINT) and self._parse_field()
            )
        )

    def _parse_json_exists(self) -> exp.JSONExists:
        this = self._parse_format_json(self._parse_bitwise())
        self._match(TokenType.COMMA)
        return self.expression(
            exp.JSONExists(
                this=this,
                path=self.dialect.to_json_path(self._parse_bitwise()),
                passing=self._match_text_seq("PASSING")
                and self._parse_csv(lambda: self._parse_alias(self._parse_bitwise())),
                on_condition=self._parse_on_condition(),
            )
        )

    def _parse_into(self) -> exp.Into | None:
        # https://docs.oracle.com/en/database/oracle/oracle-database/19/lnpls/SELECT-INTO-statement.html
        bulk_collect = self._match(TokenType.BULK_COLLECT_INTO)
        if not bulk_collect and not self._match(TokenType.INTO):
            return None

        index = self._index

        expressions = self._parse_expressions()
        if len(expressions) == 1:
            self._retreat(index)
            self._match(TokenType.TABLE)
            return self.expression(
                exp.Into(this=self._parse_table(schema=True), bulk_collect=bulk_collect)
            )

        return self.expression(exp.Into(bulk_collect=bulk_collect, expressions=expressions))

    def _parse_connect_with_prior(self):
        return self._parse_assignment()

    def _parse_column_ops(self, this: exp.Expr | None) -> exp.Expr | None:
        this = super()._parse_column_ops(this)

        if not this:
            return this

        index = self._index

        # https://docs.oracle.com/en/database/oracle/oracle-database/26/sqlrf/Interval-Exprs.html
        interval_span = self._try_parse(lambda: self._parse_interval_span(this))
        if interval_span and isinstance(interval_span.args.get("unit"), exp.IntervalSpan):
            return interval_span

        self._retreat(index)
        return this

    def _parse_insert_table(self) -> exp.Expr | None:
        # Oracle does not use AS for INSERT INTO alias
        # https://docs.oracle.com/en/database/oracle/oracle-database/18/sqlrf/INSERT.html
        # Parse table parts without schema to avoid parsing the alias with its columns
        this = self._parse_table_parts(schema=True)

        if isinstance(this, exp.Table):
            alias_name = self._parse_id_var(any_token=False)
            if alias_name:
                this.set("alias", exp.TableAlias(this=alias_name))

            this.set("partition", self._parse_partition())

            # Now parse the schema (column list) if present
            return self._parse_schema(this=this)

        return this



from __future__ import annotations

import typing as t

from sqlglot import exp, transforms
from sqlglot.dialects.dialect import (
    bracket_to_element_at_sql,
    is_parse_json,
    rename_func,
    timestamptrunc_sql,
    weekstart_unit_to_str,
)
from sqlglot.generators.hive import HIVE_DATE_FORMAT, HiveGenerator, HIVE_TS_OR_DS_EXPRESSIONS
from sqlglot.transforms import (
    preprocess,
    remove_unique_constraints,
    ctas_with_tmp_tables_to_create_tmp_view,
    move_schema_columns_to_partitioned_by,
)


def _json_format_sql(self: Spark2Generator, expression: exp.JSONFormat) -> str:
    this = expression.this

    if is_parse_json(this):
        if this.this.is_string:
            # Since FROM_JSON requires a nested type, we always wrap the json string with
            # an array to ensure that "naked" strings like "'a'" will be handled correctly
            wrapped_json = exp.Literal.string(f"[{this.this.name}]")

            from_json = self.func(
                "FROM_JSON", wrapped_json, self.func("SCHEMA_OF_JSON", wrapped_json)
            )
            to_json = self.func("TO_JSON", from_json)

            # This strips the [, ] delimiters of the dummy array printed by TO_JSON
            return self.func("REGEXP_EXTRACT", to_json, "'^.(.*).$'", "1")
        return self.sql(this)

    return self.func("TO_JSON", this, expression.args.get("options"))


def _map_sql(self: Spark2Generator, expression: exp.Map) -> str:
    keys = expression.args.get("keys")
    values = expression.args.get("values")

    if not keys or not values:
        return self.func("MAP")

    return self.func("MAP_FROM_ARRAYS", keys, values)


def _str_to_date(self: Spark2Generator, expression: exp.StrToDate) -> str:
    time_format = self.format_time(expression)
    if time_format == HIVE_DATE_FORMAT:
        return self.func("TO_DATE", expression.this)
    return self.func("TO_DATE", expression.this, time_format)


def _unix_to_time_sql(self: Spark2Generator, expression: exp.UnixToTime) -> str:
    scale = expression.args.get("scale")
    timestamp = expression.this

    if scale is None:
        return self.sql(exp.cast(exp.func("from_unixtime", timestamp), exp.DType.TIMESTAMP))
    if scale == exp.UnixToTime.SECONDS:
        return self.func("TIMESTAMP_SECONDS", timestamp)
    if scale == exp.UnixToTime.MILLIS:
        return self.func("TIMESTAMP_MILLIS", timestamp)
    if scale == exp.UnixToTime.MICROS:
        return self.func("TIMESTAMP_MICROS", timestamp)

    unix_seconds = exp.Div(this=timestamp, expression=exp.func("POW", 10, scale))
    return self.func("TIMESTAMP_SECONDS", unix_seconds)


def _unalias_pivot(expression: exp.Expr) -> exp.Expr:
    """
    Spark doesn't allow PIVOT aliases, so we need to remove them and possibly wrap a
    pivoted source in a subquery with the same alias to preserve the query's semantics.

    Example:
        >>> from sqlglot import parse_one
        >>> expr = parse_one("SELECT piv.x FROM tbl PIVOT (SUM(a) FOR b IN ('x')) piv")
        >>> print(_unalias_pivot(expr).sql(dialect="spark"))
        SELECT piv.x FROM (SELECT * FROM tbl PIVOT(SUM(a) FOR b IN ('x'))) AS piv
    """
    if isinstance(expression, exp.From) and expression.this.args.get("pivots"):
        pivot = expression.this.args["pivots"][0]
        if pivot.alias:
            alias = pivot.args["alias"].pop()
            return exp.From(
                this=expression.this.replace(
                    exp.select("*")
                    .from_(expression.this.copy(), copy=False)
                    .subquery(alias=alias, copy=False)
                )
            )

    return expression


def temporary_storage_provider(expression: exp.Expr) -> exp.Expr:
    # spark2, spark, Databricks require a storage provider for temporary tables
    provider = exp.FileFormatProperty(this=exp.Literal.string("parquet"))
    expression.args["properties"].append("expressions", provider)
    return expression


class Spark2Generator(HiveGenerator):
    QUERY_HINTS = True
    NVL2_SUPPORTED = True
    CAN_IMPLEMENT_ARRAY_ANY = True
    ALTER_SET_TYPE = "TYPE"
    PARSE_JSON_NAME: str | None = None

    PROPERTIES_LOCATION = {
        **HiveGenerator.PROPERTIES_LOCATION,
        exp.EngineProperty: exp.Properties.Location.UNSUPPORTED,
        exp.AutoIncrementProperty: exp.Properties.Location.UNSUPPORTED,
        exp.CharacterSetProperty: exp.Properties.Location.UNSUPPORTED,
        exp.CollateProperty: exp.Properties.Location.UNSUPPORTED,
    }

    TS_OR_DS_EXPRESSIONS: t.ClassVar = (
        *HIVE_TS_OR_DS_EXPRESSIONS,
        exp.DayOfMonth,
        exp.DayOfWeek,
        exp.DayOfYear,
        exp.WeekOfYear,
    )

    TRANSFORMS = {
        k: v
        for k, v in {
            **HiveGenerator.TRANSFORMS,
            exp.ApproxDistinct: rename_func("APPROX_COUNT_DISTINCT"),
            exp.ArraySum: lambda self, e: (
                f"AGGREGATE({self.sql(e, 'this')}, 0, (acc, x) -> acc + x, acc -> acc)"
            ),
            exp.ArrayToString: rename_func("ARRAY_JOIN"),
            exp.ArraySlice: rename_func("SLICE"),
            exp.AtTimeZone: lambda self, e: self.func(
                "FROM_UTC_TIMESTAMP", e.this, e.args.get("zone")
            ),
            exp.BitwiseLeftShift: rename_func("SHIFTLEFT"),
            exp.BitwiseRightShift: rename_func("SHIFTRIGHT"),
            exp.Create: preprocess(
                [
                    remove_unique_constraints,
                    lambda e: ctas_with_tmp_tables_to_create_tmp_view(
                        e, temporary_storage_provider
                    ),
                    move_schema_columns_to_partitioned_by,
                ]
            ),
            exp.DateFromParts: rename_func("MAKE_DATE"),
            exp.DateTrunc: lambda self, e: self.func(
                "TRUNC", e.this, weekstart_unit_to_str(self, e)
            ),
            exp.DayOfMonth: rename_func("DAYOFMONTH"),
            exp.DayOfWeek: rename_func("DAYOFWEEK"),
            # (DAY_OF_WEEK(datetime) % 7) + 1 is equivalent to DAYOFWEEK_ISO(datetime)
            exp.DayOfWeekIso: lambda self, e: f"(({self.func('DAYOFWEEK', e.this)} % 7) + 1)",
            exp.DayOfYear: rename_func("DAYOFYEAR"),
            exp.Format: rename_func("FORMAT_STRING"),
            exp.From: transforms.preprocess([_unalias_pivot]),
            exp.FromTimeZone: lambda self, e: self.func(
                "TO_UTC_TIMESTAMP", e.this, e.args.get("zone")
            ),
            exp.JSONFormat: _json_format_sql,
            exp.LogicalAnd: rename_func("BOOL_AND"),
            exp.LogicalOr: rename_func("BOOL_OR"),
            exp.Map: _map_sql,
            exp.Pivot: transforms.preprocess([transforms.unqualify_pivot_fields]),
            exp.Reduce: rename_func("AGGREGATE"),
            exp.RegexpReplace: lambda self, e: self.func(
                "REGEXP_REPLACE",
                e.this,
                e.expression,
                e.args["replacement"],
                e.args.get("position"),
            ),
            exp.Select: transforms.preprocess(
                [
                    transforms.eliminate_qualify,
                    transforms.eliminate_distinct_on,
                    transforms.unnest_to_explode,
                    transforms.any_to_exists,
                ]
            ),
            exp.SHA2Digest: lambda self, e: self.func(
                "SHA2", e.this, e.args.get("length") or exp.Literal.number(256)
            ),
            exp.StrToDate: _str_to_date,
            exp.StrToTime: lambda self, e: self.func("TO_TIMESTAMP", e.this, self.format_time(e)),
            exp.TimestampTrunc: timestamptrunc_sql(),
            exp.UnixToTime: _unix_to_time_sql,
            exp.VariancePop: rename_func("VAR_POP"),
            exp.WeekOfYear: rename_func("WEEKOFYEAR"),
            exp.WithinGroup: transforms.preprocess(
                [transforms.remove_within_group_for_percentiles]
            ),
            exp.ArraySort: None,
            exp.ILike: None,
            exp.Left: None,
            exp.MonthsBetween: None,
            exp.Right: None,
        }.items()
        if v is not None
    }

    WRAP_DERIVED_VALUES = False
    CREATE_FUNCTION_RETURN_AS = False

    def struct_sql(self, expression: exp.Struct) -> str:
        from sqlglot.generator import Generator

        return Generator.struct_sql(self, expression)

    def cast_sql(self, expression: exp.Cast, safe_prefix: str | None = None) -> str:
        arg = expression.this
        is_json_extract = isinstance(
            arg, (exp.JSONExtract, exp.JSONExtractScalar)
        ) and not arg.args.get("variant_extract")

        # We can't use a non-nested type (eg. STRING) as a schema
        if expression.to.args.get("nested") and (is_parse_json(arg) or is_json_extract):
            schema = f"'{self.sql(expression, 'to')}'"
            return self.func("FROM_JSON", arg if is_json_extract else arg.this, schema)

        if is_parse_json(expression):
            return self.func("TO_JSON", arg)

        return super(HiveGenerator, self).cast_sql(expression, safe_prefix=safe_prefix)

    def fileformatproperty_sql(self, expression: exp.FileFormatProperty) -> str:
        if expression.args.get("hive_format"):
            return super().fileformatproperty_sql(expression)

        return f"USING {expression.name.upper()}"

    def altercolumn_sql(self, expression: exp.AlterColumn) -> str:
        if expression.args.get("exists"):
            self.unsupported("ALTER COLUMN IF EXISTS is not supported by this dialect")

        this = self.sql(expression, "this")
        new_name = self.sql(expression, "rename_to") or this
        comment = self.sql(expression, "comment")
        if new_name == this:
            if comment:
                return f"ALTER COLUMN {this} COMMENT {comment}"
            return super(HiveGenerator, self).altercolumn_sql(expression)
        return f"RENAME COLUMN {this} TO {new_name}"

    def renamecolumn_sql(self, expression: exp.RenameColumn) -> str:
        return super(HiveGenerator, self).renamecolumn_sql(expression)

    def bracket_sql(self, expression: exp.Bracket) -> str:
        if expression.args.get("safe") is False:
            return bracket_to_element_at_sql(self, expression)

        return super().bracket_sql(expression)



from __future__ import annotations

import typing as t

from sqlglot import exp, parser
from sqlglot.dialects.dialect import (
    DialectType,
    binary_from_function,
    build_formatted_time,
    build_json_extract_path,
    build_timestamp_trunc,
)
from sqlglot.helper import is_int, seq_get
from sqlglot.parser import binary_range_parser
from sqlglot.tokens import TokenType

if t.TYPE_CHECKING:
    from sqlglot.dialects.dialect import Dialect


def _build_generate_series(args: list) -> exp.ExplodingGenerateSeries:
    # The goal is to convert step values like '1 day' or INTERVAL '1 day' into INTERVAL '1' day
    # Note: postgres allows calls with just two arguments -- the "step" argument defaults to 1
    step = seq_get(args, 2)
    if step is not None:
        if step.is_string:
            args[2] = exp.to_interval(step.this)
        elif isinstance(step, exp.Interval) and not step.args.get("unit"):
            args[2] = exp.to_interval(step.this.this)

    return exp.ExplodingGenerateSeries.from_arg_list(args)


def _build_to_timestamp(args: list, dialect: Dialect) -> exp.UnixToTime | exp.StrToTime:
    # TO_TIMESTAMP accepts either a single double argument or (text, text)
    if len(args) == 1:
        # https://www.postgresql.org/docs/current/functions-datetime.html#FUNCTIONS-DATETIME-TABLE
        return exp.UnixToTime.from_arg_list(args)

    # https://www.postgresql.org/docs/current/functions-formatting.html
    return build_formatted_time(exp.StrToTime)(args, dialect)


def _build_regexp_replace(args: list, dialect: DialectType = None) -> exp.RegexpReplace:
    # The signature of REGEXP_REPLACE is:
    # regexp_replace(source, pattern, replacement [, start [, N ]] [, flags ])
    #
    # Any one of `start`, `N` and `flags` can be column references, meaning that
    # unless we can statically see that the last argument is a non-integer string
    # (eg. not '0'), then it's not possible to construct the correct AST
    regexp_replace = None
    if len(args) > 3:
        last = args[-1]
        if not is_int(last.name):
            if not last.type or last.is_type(exp.DType.UNKNOWN, exp.DType.NULL):
                from sqlglot.optimizer.annotate_types import annotate_types

                last = annotate_types(last, dialect=dialect)

            if last.is_type(*exp.DataType.TEXT_TYPES):
                regexp_replace = exp.RegexpReplace.from_arg_list(args[:-1])
                regexp_replace.set("modifiers", last)

    regexp_replace = regexp_replace or exp.RegexpReplace.from_arg_list(args)
    regexp_replace.set("single_replace", True)
    return regexp_replace


def _build_levenshtein_less_equal(args: list) -> exp.Levenshtein:
    # Postgres has two signatures for levenshtein_less_equal function, but in both cases
    # max_dist is the last argument
    # levenshtein_less_equal(source, target, ins_cost, del_cost, sub_cost, max_d)
    # levenshtein_less_equal(source, target, max_d)
    max_dist = args.pop()

    return exp.Levenshtein(
        this=seq_get(args, 0),
        expression=seq_get(args, 1),
        ins_cost=seq_get(args, 2),
        del_cost=seq_get(args, 3),
        sub_cost=seq_get(args, 4),
        max_dist=max_dist,
    )


class PostgresParser(parser.Parser):
    SUPPORTS_OMITTED_INTERVAL_SPAN_UNIT = True

    PROPERTY_PARSERS = {
        **{k: v for k, v in parser.Parser.PROPERTY_PARSERS.items() if k != "INPUT"},
        "SET": lambda self: self.expression(exp.SetConfigProperty(this=self._parse_set())),
    }

    PLACEHOLDER_PARSERS = {
        **parser.Parser.PLACEHOLDER_PARSERS,
        TokenType.PLACEHOLDER: lambda self: self.expression(exp.Placeholder(jdbc=True)),
        TokenType.MOD: lambda self: self._parse_query_parameter(),
    }

    FUNCTIONS = {
        **parser.Parser.FUNCTIONS,
        "ARRAY_PREPEND": lambda args: exp.ArrayPrepend(
            this=seq_get(args, 1), expression=seq_get(args, 0)
        ),
        "BIT_AND": exp.BitwiseAndAgg.from_arg_list,
        "BIT_OR": exp.BitwiseOrAgg.from_arg_list,
        "BIT_XOR": exp.BitwiseXorAgg.from_arg_list,
        "VERSION": exp.CurrentVersion.from_arg_list,
        "DATE_TRUNC": build_timestamp_trunc,
        "DIV": lambda args: exp.cast(binary_from_function(exp.IntDiv)(args), exp.DType.DECIMAL),
        "GENERATE_SERIES": _build_generate_series,
        "GET_BIT": lambda args: exp.Getbit(
            this=seq_get(args, 0), expression=seq_get(args, 1), zero_is_msb=True
        ),
        "JSON_EXTRACT_PATH": build_json_extract_path(exp.JSONExtract),
        "JSON_EXTRACT_PATH_TEXT": build_json_extract_path(exp.JSONExtractScalar),
        "LENGTH": lambda args: exp.Length(this=seq_get(args, 0), encoding=seq_get(args, 1)),
        "MAKE_TIME": exp.TimeFromParts.from_arg_list,
        "MAKE_TIMESTAMP": exp.TimestampFromParts.from_arg_list,
        "NOW": exp.CurrentTimestamp.from_arg_list,
        "REGEXP_REPLACE": _build_regexp_replace,
        "TO_CHAR": build_formatted_time(exp.TimeToStr),
        "TO_DATE": build_formatted_time(exp.StrToDate),
        "TO_TIMESTAMP": _build_to_timestamp,
        "UNNEST": exp.Explode.from_arg_list,
        "SHA256": lambda args: exp.SHA2(this=seq_get(args, 0), length=exp.Literal.number(256)),
        "SHA384": lambda args: exp.SHA2(this=seq_get(args, 0), length=exp.Literal.number(384)),
        "SHA512": lambda args: exp.SHA2(this=seq_get(args, 0), length=exp.Literal.number(512)),
        "LEVENSHTEIN_LESS_EQUAL": _build_levenshtein_less_equal,
        "JSON_OBJECT_AGG": lambda args: exp.JSONObjectAgg(expressions=args),
        "JSONB_OBJECT_AGG": exp.JSONBObjectAgg.from_arg_list,
        "WIDTH_BUCKET": lambda args: (
            exp.WidthBucket(this=seq_get(args, 0), threshold=seq_get(args, 1))
            if len(args) == 2
            else exp.WidthBucket.from_arg_list(args)
        ),
        "UUID": lambda args: exp.cast(args[0], exp.DType.UUID) if args else exp.Uuid(),
    }

    NO_PAREN_FUNCTION_PARSERS = {
        **parser.Parser.NO_PAREN_FUNCTION_PARSERS,
        "VARIADIC": lambda self: self.expression(exp.Variadic(this=self._parse_bitwise())),
    }

    NO_PAREN_FUNCTIONS = {
        **parser.Parser.NO_PAREN_FUNCTIONS,
        TokenType.LOCALTIME: exp.Localtime,
        TokenType.LOCALTIMESTAMP: exp.Localtimestamp,
        TokenType.CURRENT_CATALOG: exp.CurrentCatalog,
        TokenType.SESSION_USER: exp.SessionUser,
        TokenType.CURRENT_SCHEMA: exp.CurrentSchema,
    }

    FUNCTION_PARSERS = {
        **parser.Parser.FUNCTION_PARSERS,
        "DATE_PART": lambda self: self._parse_date_part(),
        "JSON_AGG": lambda self: self.expression(
            exp.JSONArrayAgg(this=self._parse_lambda(), order=self._parse_order())
        ),
        "JSONB_EXISTS": lambda self: self._parse_jsonb_exists(),
    }

    BITWISE = {
        **parser.Parser.BITWISE,
        TokenType.HASH: exp.BitwiseXor,
    }

    EXPONENT = {
        TokenType.CARET: exp.Pow,
    }

    RANGE_PARSERS = {
        **parser.Parser.RANGE_PARSERS,
        TokenType.DAMP: binary_range_parser(exp.ArrayOverlaps),
        TokenType.DAT: lambda self, this: self.expression(
            exp.MatchAgainst(this=self._parse_bitwise(), expressions=[this])
        ),
    }

    STATEMENT_PARSERS = {
        **parser.Parser.STATEMENT_PARSERS,
        TokenType.END: lambda self: self._parse_commit_or_rollback(),
    }

    UNARY_PARSERS = {
        **parser.Parser.UNARY_PARSERS,
        # The `~` token is remapped from TILDE to RLIKE in Postgres due to the binary REGEXP LIKE operator
        TokenType.RLIKE: lambda self: self.expression(exp.BitwiseNot(this=self._parse_unary())),
    }

    JSON_ARROWS_REQUIRE_JSON_TYPE = True

    COLUMN_OPERATORS = {
        k: v
        for k, v in parser.Parser.COLUMN_OPERATORS.items()
        if k
        not in (
            TokenType.ARROW,
            TokenType.DARROW,
            TokenType.HASH_ARROW,
            TokenType.DHASH_ARROW,
            TokenType.PLACEHOLDER,
        )
    }

    JSON_OPERATORS = {
        TokenType.ARROW: lambda self, this, path: self.validate_expression(
            build_json_extract_path(
                exp.JSONExtract, arrow_req_json_type=self.JSON_ARROWS_REQUIRE_JSON_TYPE
            )([this, path])
        ),
        TokenType.DARROW: lambda self, this, path: self.validate_expression(
            build_json_extract_path(
                exp.JSONExtractScalar, arrow_req_json_type=self.JSON_ARROWS_REQUIRE_JSON_TYPE
            )([this, path])
        ),
        TokenType.HASH_ARROW: parser.build_jsonb_extract,
        TokenType.DHASH_ARROW: parser.build_jsonb_extract_scalar,
        TokenType.PLACEHOLDER: parser.build_jsonb_contains,
    }

    ARG_MODE_TOKENS: t.ClassVar = {TokenType.IN, TokenType.OUT, TokenType.INOUT, TokenType.VARIADIC}

    def _parse_parameter_mode(self) -> TokenType | None:
        """
        Parse PostgreSQL function parameter mode (IN, OUT, INOUT, VARIADIC).

        Disambiguates between mode keywords and identifiers with the same name:
        - MODE TYPE      -> keyword is identifier (e.g., "out INT")
        - MODE NAME TYPE -> keyword is mode (e.g., "OUT x INT")

        Returns:
            Mode token type if current token is a mode keyword, None otherwise.
        """
        if not self._match_set(self.ARG_MODE_TOKENS, advance=False) or not self._next:
            return None

        mode_token = self._curr

        # Check Pattern 1: MODE TYPE
        # Try parsing next token as a built-in type (not UDT)
        # If successful, the keyword is an identifier, not a mode
        is_followed_by_builtin_type = self._try_parse(
            lambda: (
                self._advance()  # type: ignore
                or self._parse_types(check_func=False, allow_identifiers=False)
            ),
            retreat=True,
        )
        if is_followed_by_builtin_type:
            return None  # Pattern: "out INT" -> out is parameter name

        # Check Pattern 2: MODE NAME TYPE
        # If next token is an identifier, check if there's a type after it
        # The type can be built-in or user-defined (allow_identifiers=True)
        if self._next.token_type not in self.ID_VAR_TOKENS:
            return None

        is_followed_by_any_type = self._try_parse(
            lambda: (
                self._advance(2)  # type: ignore
                or self._parse_types(check_func=False, allow_identifiers=True)
            ),
            retreat=True,
        )

        if is_followed_by_any_type:
            return mode_token.token_type  # Pattern: "OUT x INT" -> OUT is mode

        return None

    def _create_mode_constraint(self, param_mode: TokenType) -> exp.InOutColumnConstraint:
        """
        Create parameter mode constraint for function parameters.

        Args:
            param_mode: The parameter mode token (IN, OUT, INOUT, or VARIADIC).

        Returns:
            InOutColumnConstraint expression representing the parameter mode.
        """
        return self.expression(
            exp.InOutColumnConstraint(
                input_=(param_mode in {TokenType.IN, TokenType.INOUT}),
                output=(param_mode in {TokenType.OUT, TokenType.INOUT}),
                variadic=(param_mode == TokenType.VARIADIC),
            )
        )

    def _parse_function_parameter(self) -> exp.Expr | None:
        param_mode = self._parse_parameter_mode()

        if param_mode:
            self._advance()

        # Parse parameter name and type
        param_name = self._parse_id_var()
        column_def = self._parse_column_def(this=param_name, computed_column=False)

        # Attach mode as constraint
        if param_mode and column_def:
            constraint = self._create_mode_constraint(param_mode)
            if not column_def.args.get("constraints"):
                column_def.set("constraints", [])
            column_def.args["constraints"].insert(0, constraint)

        return column_def

    def _parse_query_parameter(self) -> exp.Expr | None:
        this = (
            self._parse_wrapped(self._parse_id_var)
            if self._match(TokenType.L_PAREN, advance=False)
            else None
        )
        self._match_text_seq("S")
        return self.expression(exp.Placeholder(this=this))

    def _parse_date_part(self) -> exp.Expr:
        part = self._parse_type()
        self._match(TokenType.COMMA)
        value = self._parse_bitwise()

        if part and isinstance(part, (exp.Column, exp.Literal)):
            part = exp.var(part.name)

        return self.expression(exp.Extract(this=part, expression=value))

    def _parse_unique_key(self) -> exp.Expr | None:
        return None

    def _parse_jsonb_exists(self) -> exp.JSONBExists:
        return self.expression(
            exp.JSONBExists(
                this=self._parse_bitwise(),
                path=self._match(TokenType.COMMA)
                and self.dialect.to_json_path(self._parse_bitwise()),
            )
        )

    def _parse_generated_as_identity(
        self,
    ) -> (
        exp.GeneratedAsIdentityColumnConstraint
        | exp.ComputedColumnConstraint
        | exp.GeneratedAsRowColumnConstraint
    ):
        this = super()._parse_generated_as_identity()

        if self._match_text_seq("STORED"):
            this = self.expression(exp.ComputedColumnConstraint(this=this.expression))

        return this

    def _parse_user_defined_type(self, identifier: exp.Identifier) -> exp.Expr | None:
        udt_type: exp.Identifier | exp.Dot = identifier

        while self._match(TokenType.DOT):
            part = self._parse_id_var()
            if part:
                udt_type = exp.Dot(this=udt_type, expression=part)

        return exp.DataType.build(udt_type, udt=True)



import os

INTEGRATION_TEST_DIR = os.path.join(
    os.path.dirname(__file__),
    "..",
    "sqlglot-integration-tests",
    "tests",
    "sqlglot",
)


def load_tests(loader, suite, pattern):
    if os.path.isdir(INTEGRATION_TEST_DIR):
        suite.addTests(loader.discover(INTEGRATION_TEST_DIR, pattern="test*.py"))
    return suite



"""sqlglot expressions - date, time, and timestamp functions."""

from __future__ import annotations

import typing as t

from sqlglot.expressions.core import (
    Expression,
    Func,
    TimeUnit,
    IntervalOp,
    Literal,
    Column,
    TIMESTAMP_PARTS,
)
from sqlglot.expressions.datatypes import DataType, DType

if t.TYPE_CHECKING:
    from sqlglot.expressions.core import Expr, Neg


# Current date/time


class CurrentDate(Expression, Func):
    arg_types = {"this": False}


class CurrentDatetime(Expression, Func):
    arg_types = {"this": False}


class CurrentTime(Expression, Func):
    arg_types = {"this": False}


class CurrentTimestamp(Expression, Func):
    arg_types = {"this": False, "sysdate": False}


class CurrentTimestampLTZ(Expression, Func):
    arg_types = {}


class CurrentTimezone(Expression, Func):
    arg_types = {}


class Localtime(Expression, Func):
    arg_types = {"this": False}


class Localtimestamp(Expression, Func):
    arg_types = {"this": False}


class Systimestamp(Expression, Func):
    arg_types = {"this": False}


class UtcDate(Expression, Func):
    arg_types = {}


class UtcTime(Expression, Func):
    arg_types = {"this": False}


class UtcTimestamp(Expression, Func):
    arg_types = {"this": False}


# Date arithmetic


class AddMonths(Expression, Func):
    arg_types = {"this": True, "expression": True, "preserve_end_of_month": False}


class DateAdd(Expression, Func, IntervalOp):
    arg_types = {"this": True, "expression": True, "unit": False}


class DateBin(Expression, Func, IntervalOp):
    arg_types = {"this": True, "expression": True, "unit": False, "zone": False, "origin": False}


class DateDiff(Expression, Func, TimeUnit):
    _sql_names = ["DATEDIFF", "DATE_DIFF"]
    arg_types = {
        "this": True,
        "expression": True,
        "unit": False,
        "zone": False,
        "big_int": False,
        "date_part_boundary": False,
    }


class DateSub(Expression, Func, IntervalOp):
    arg_types = {"this": True, "expression": True, "unit": False}


class DatetimeAdd(Expression, Func, IntervalOp):
    arg_types = {"this": True, "expression": True, "unit": False}


class DatetimeDiff(Expression, Func, TimeUnit):
    arg_types = {"this": True, "expression": True, "unit": False, "date_part_boundary": False}


class DatetimeSub(Expression, Func, IntervalOp):
    arg_types = {"this": True, "expression": True, "unit": False}


class MonthsBetween(Expression, Func):
    arg_types = {"this": True, "expression": True, "roundoff": False}


class TimeAdd(Expression, Func, TimeUnit):
    arg_types = {"this": True, "expression": True, "unit": False}


class TimeDiff(Expression, Func, TimeUnit):
    arg_types = {"this": True, "expression": True, "unit": False}


class TimeSub(Expression, Func, TimeUnit):
    arg_types = {"this": True, "expression": True, "unit": False}


class TimestampAdd(Expression, Func, TimeUnit):
    arg_types = {"this": True, "expression": True, "unit": False}


class TimestampDiff(Expression, Func, TimeUnit):
    _sql_names = ["TIMESTAMPDIFF", "TIMESTAMP_DIFF"]
    arg_types = {"this": True, "expression": True, "unit": False}


class TimestampSub(Expression, Func, TimeUnit):
    arg_types = {"this": True, "expression": True, "unit": False}


class TsOrDsAdd(Expression, Func, TimeUnit):
    # return_type is used to correctly cast the arguments of this expression when transpiling it
    arg_types = {"this": True, "expression": True, "unit": False, "return_type": False}

    @property
    def return_type(self) -> DataType:
        return DataType.build(self.args.get("return_type") or DType.DATE)


class TsOrDsDiff(Expression, Func, TimeUnit):
    arg_types = {"this": True, "expression": True, "unit": False}


# Truncation


class DatetimeTrunc(Expression, Func, TimeUnit):
    arg_types = {"this": True, "unit": True, "zone": False}


class DateTrunc(Expression, Func):
    arg_types = {"unit": True, "this": True, "zone": False, "input_type_preserved": False}

    def __init__(self, **args):
        # Across most dialects it's safe to unabbreviate the unit (e.g. 'Q' -> 'QUARTER') except Oracle
        # https://docs.oracle.com/en/database/oracle/oracle-database/21/sqlrf/ROUND-and-TRUNC-Date-Functions.html
        unabbreviate = args.pop("unabbreviate", True)

        unit = args.get("unit")
        if isinstance(unit, TimeUnit.VAR_LIKE) and not (
            isinstance(unit, Column) and len(unit.parts) != 1
        ):
            unit_name = unit.name.upper()
            if unabbreviate and unit_name in TimeUnit.UNABBREVIATED_UNIT_NAME:
                unit_name = TimeUnit.UNABBREVIATED_UNIT_NAME[unit_name]

            args["unit"] = Literal.string(unit_name)

        super().__init__(**args)

    @property
    def unit(self) -> Expr:
        return self.args["unit"]


class TimestampTrunc(Expression, Func, TimeUnit):
    arg_types = {"this": True, "unit": True, "zone": False, "input_type_preserved": False}


class TimeSlice(Expression, Func, TimeUnit):
    arg_types = {"this": True, "expression": True, "unit": True, "kind": False}


class TimeTrunc(Expression, Func, TimeUnit):
    arg_types = {"this": True, "unit": True, "zone": False}


# Date/time extraction


class Day(Expression, Func):
    pass


class DayOfMonth(Expression, Func):
    _sql_names = ["DAY_OF_MONTH", "DAYOFMONTH"]


class DayOfWeek(Expression, Func):
    _sql_names = ["DAY_OF_WEEK", "DAYOFWEEK"]


class DayOfWeekIso(Expression, Func):
    _sql_names = ["DAYOFWEEK_ISO", "ISODOW"]


class DayOfYear(Expression, Func):
    _sql_names = ["DAY_OF_YEAR", "DAYOFYEAR"]


class Dayname(Expression, Func):
    arg_types = {"this": True, "abbreviated": False}


class Extract(Expression, Func):
    arg_types = {"this": True, "expression": True}


class GetExtract(Expression, Func):
    arg_types = {"this": True, "expression": True}


class Hour(Expression, Func):
    pass


class Minute(Expression, Func):
    pass


class Month(Expression, Func):
    pass


class Monthname(Expression, Func):
    arg_types = {"this": True, "abbreviated": False}


class Quarter(Expression, Func):
    pass


class Second(Expression, Func):
    pass


class ToDays(Expression, Func):
    pass


class Week(Expression, Func):
    arg_types = {"this": True, "mode": False}


class WeekOfYear(Expression, Func):
    _sql_names = ["WEEK_OF_YEAR", "WEEKOFYEAR"]


class Year(Expression, Func):
    pass


class YearOfWeek(Expression, Func):
    _sql_names = ["YEAR_OF_WEEK", "YEAROFWEEK"]


class YearOfWeekIso(Expression, Func):
    _sql_names = ["YEAR_OF_WEEK_ISO", "YEAROFWEEKISO"]


# Date/time construction


class Date(Expression, Func):
    arg_types = {"this": False, "zone": False, "expressions": False}
    is_var_len_args = True


class DateFromParts(Expression, Func):
    _sql_names = ["DATE_FROM_PARTS", "DATEFROMPARTS"]
    arg_types = {"year": True, "month": False, "day": False, "allow_overflow": False}


class DateFromUnixDate(Expression, Func):
    pass


class Datetime(Expression, Func):
    arg_types = {"this": True, "expression": False}


class GapFill(Expression, Func):
    arg_types = {
        "this": True,
        "ts_column": True,
        "bucket_width": True,
        "partitioning_columns": False,
        "value_columns": False,
        "origin": False,
        "ignore_nulls": False,
    }


class GenerateDateArray(Expression, Func):
    arg_types = {"start": True, "end": True, "step": False}


class GenerateTimestampArray(Expression, Func):
    arg_types = {"start": True, "end": True, "step": True}


class JustifyDays(Expression, Func):
    pass


class JustifyHours(Expression, Func):
    pass


class JustifyInterval(Expression, Func):
    pass


class LastDay(Expression, Func, TimeUnit):
    _sql_names = ["LAST_DAY", "LAST_DAY_OF_MONTH"]
    arg_types = {"this": True, "unit": False}


class MakeInterval(Expression, Func):
    arg_types = {
        "year": False,
        "month": False,
        "week": False,
        "day": False,
        "hour": False,
        "minute": False,
        "second": False,
    }


class NextDay(Expression, Func):
    arg_types = {"this": True, "expression": True}


class PreviousDay(Expression, Func):
    arg_types = {"this": True, "expression": True}


class Time(Expression, Func):
    arg_types = {"this": False, "zone": False}


class TimeFromParts(Expression, Func):
    _sql_names = ["TIME_FROM_PARTS", "TIMEFROMPARTS"]
    arg_types = {
        "hour": True,
        "min": True,
        "sec": True,
        "nano": False,
        "fractions": False,
        "precision": False,
        "overflow": False,
    }


class Timestamp(Expression, Func):
    arg_types = {"this": False, "zone": False, "with_tz": False}


class TimestampFromParts(Expression, Func):
    _sql_names = ["TIMESTAMP_FROM_PARTS", "TIMESTAMPFROMPARTS"]
    arg_types = {
        **TIMESTAMP_PARTS,
        "zone": False,
        "milli": False,
        "this": False,
        "expression": False,
    }


class TimestampLtzFromParts(Expression, Func):
    _sql_names = ["TIMESTAMP_LTZ_FROM_PARTS", "TIMESTAMPLTZFROMPARTS"]
    arg_types = TIMESTAMP_PARTS.copy()


class TimestampTzFromParts(Expression, Func):
    _sql_names = ["TIMESTAMP_TZ_FROM_PARTS", "TIMESTAMPTZFROMPARTS"]
    arg_types = {
        **TIMESTAMP_PARTS,
        "zone": False,
    }


# Date/time conversion


class ConvertTimezone(Expression, Func):
    arg_types = {
        "source_tz": False,
        "target_tz": True,
        "timestamp": True,
        "options": False,
    }


class DateStrToDate(Expression, Func):
    pass


class DateToDateStr(Expression, Func):
    pass


class DateToDi(Expression, Func):
    pass


class DiToDate(Expression, Func):
    pass


class FromISO8601Timestamp(Expression, Func):
    _sql_names = ["FROM_ISO8601_TIMESTAMP"]


class FromISO8601Date(Expression, Func):
    _sql_names = ["FROM_ISO8601_DATE"]


class FromISO8601TimestampNanos(Expression, Func):
    _sql_names = ["FROM_ISO8601_TIMESTAMP_NANOS"]


class ParseDatetime(Expression, Func):
    arg_types = {"this": True, "format": False, "zone": False, "default_year": False}


class ParseTime(Expression, Func):
    arg_types = {"this": True, "format": True}


class StrToDate(Expression, Func):
    arg_types = {"this": True, "format": False, "safe": False, "default_year": False}


class StrToTime(Expression, Func):
    arg_types = {
        "this": True,
        "format": True,
        "zone": False,
        "safe": False,
        "target_type": False,
        "default_year": False,
    }


class StrToUnix(Expression, Func):
    arg_types = {"this": False, "format": False}


class TimeStrToDate(Expression, Func):
    pass


class TimeStrToTime(Expression, Func):
    arg_types = {"this": True, "zone": False}


class TimeStrToUnix(Expression, Func):
    pass


class TimeToStr(Expression, Func):
    arg_types = {"this": True, "format": True, "culture": False, "zone": False}


class TimeToTimeStr(Expression, Func):
    pass


class TimeToUnix(Expression, Func):
    pass


class TsOrDiToDi(Expression, Func):
    pass


class TsOrDsToDate(Expression, Func):
    arg_types = {"this": True, "format": False, "safe": False, "default_date": False}


class TsOrDsToDateStr(Expression, Func):
    pass


class TsOrDsToDatetime(Expression, Func):
    pass


class TsOrDsToTime(Expression, Func):
    arg_types = {"this": True, "format": False, "safe": False}


class TsOrDsToTimestamp(Expression, Func):
    pass


class UnixDate(Expression, Func):
    pass


class UnixMicros(Expression, Func):
    pass


class UnixMillis(Expression, Func):
    pass


class UnixSeconds(Expression, Func):
    pass


class UnixToStr(Expression, Func):
    arg_types = {"this": True, "format": False}


class UnixToTime(Expression, Func):
    arg_types = {
        "this": True,
        "scale": False,
        "zone": False,
        "hours": False,
        "minutes": False,
        "format": False,
        "target_type": False,
    }

    SECONDS: t.ClassVar[Literal | Neg] = Literal.number(0)
    DECIS: t.ClassVar[Literal | Neg] = Literal.number(1)
    CENTIS: t.ClassVar[Literal | Neg] = Literal.number(2)
    MILLIS: t.ClassVar[Literal | Neg] = Literal.number(3)
    DECIMILLIS: t.ClassVar[Literal | Neg] = Literal.number(4)
    CENTIMILLIS: t.ClassVar[Literal | Neg] = Literal.number(5)
    MICROS: t.ClassVar[Literal | Neg] = Literal.number(6)
    DECIMICROS: t.ClassVar[Literal | Neg] = Literal.number(7)
    CENTIMICROS: t.ClassVar[Literal | Neg] = Literal.number(8)
    NANOS: t.ClassVar[Literal | Neg] = Literal.number(9)


class UnixToTimeStr(Expression, Func):
    pass



from tests.dialects.test_dialect import Validator


class TestDrill(Validator):
    dialect = "drill"

    def test_drill(self):
        self.validate_identity(
            "SELECT * FROM table(dfs.`test_data.xlsx`(type => 'excel', sheetName => 'secondSheet'))"
        )
        self.validate_identity(
            "SELECT * FROM (SELECT * FROM t) PIVOT(avg(c1) AS ac1 FOR c2 IN ('V' AS v))",
        )

        self.validate_all(
            "SELECT '2021-01-01' + INTERVAL 1 MONTH",
            write={
                "drill": "SELECT '2021-01-01' + INTERVAL '1' MONTH",
                "mysql": "SELECT '2021-01-01' + INTERVAL '1' MONTH",
            },
        )

    def test_ilike(self):
        self.validate_all(
            "SELECT ILIKE(x, '%y')",
            write={
                "drill": "SELECT ILIKE(x, '%y')",
                "duckdb": "SELECT x ILIKE '%y'",
                "postgres": "SELECT x ILIKE '%y'",
            },
        )
        self.validate_all(
            "SELECT NOT ILIKE(x, '%y')",
            write={
                "drill": "SELECT NOT ILIKE(x, '%y')",
                "postgres": "SELECT NOT x ILIKE '%y'",
            },
        )
        self.validate_all(
            "SELECT ILIKE(x, '%y', '#')",
            write={
                "drill": "SELECT ILIKE(x, '%y', '#')",
                "postgres": "SELECT x ILIKE '%y' ESCAPE '#'",
            },
        )
        self.validate_all(
            "NOT ILIKE(x, '%y', '#')",
            read={"postgres": "x NOT ILIKE '%y' ESCAPE '#'"},
        )
        self.validate_all(
            "ILIKE(x, 'a') OR ILIKE(x, 'b')",
            read={"snowflake": "x ILIKE ANY ('a', 'b')"},
        )
        self.validate_all(
            "ILIKE(x, 'a', '#') OR ILIKE(x, 'b', '#')",
            read={"snowflake": "x ILIKE ANY ('a', 'b') ESCAPE '#'"},
        )
        self.validate_all(
            "x LIKE 'a' OR x LIKE 'b'",
            read={"snowflake": "x LIKE ANY ('a', 'b')"},
        )

    def test_analyze(self):
        self.validate_identity("ANALYZE TABLE tbl COMPUTE STATISTICS")
        self.validate_identity("ANALYZE TABLE tbl COMPUTE STATISTICS SAMPLE 5 PERCENT")



from __future__ import annotations

import typing as t

from sqlglot import exp
from sqlglot.dialects.dialect import Dialect, DialectType
from sqlglot.optimizer.isolate_table_selects import isolate_table_selects
from sqlglot.optimizer.normalize_identifiers import normalize_identifiers
from sqlglot.optimizer.qualify_columns import (
    qualify_columns as qualify_columns_func,
    quote_identifiers as quote_identifiers_func,
    validate_qualify_columns as validate_qualify_columns_func,
)
from sqlglot.optimizer.qualify_tables import qualify_tables
from sqlglot.schema import Schema, ensure_schema
from sqlglot._typing import E


def qualify(
    expression: E,
    dialect: DialectType = None,
    db: str | None = None,
    catalog: str | None = None,
    schema: dict[str, object] | Schema | None = None,
    expand_alias_refs: bool = True,
    expand_stars: bool = True,
    infer_schema: bool | None = None,
    isolate_tables: bool = False,
    qualify_columns: bool = True,
    allow_partial_qualification: bool = False,
    validate_qualify_columns: bool = True,
    quote_identifiers: bool = True,
    identify: bool = True,
    canonicalize_table_aliases: bool = False,
    on_qualify: t.Callable[[exp.Table], None] | None = None,
    sql: str | None = None,
) -> E:
    """
    Rewrite sqlglot AST to have normalized and qualified tables and columns.

    This step is necessary for all further SQLGlot optimizations.

    Example:
        >>> import sqlglot
        >>> schema = {"tbl": {"col": "INT"}}
        >>> expression = sqlglot.parse_one("SELECT col FROM tbl")
        >>> qualify(expression, schema=schema).sql()
        'SELECT "tbl"."col" AS "col" FROM "tbl" AS "tbl"'

    Args:
        expression: Expr to qualify.
        db: Default database name for tables.
        catalog: Default catalog name for tables.
        schema: Schema to infer column names and types.
        expand_alias_refs: Whether to expand references to aliases.
        expand_stars: Whether to expand star queries. This is a necessary step
            for most of the optimizer's rules to work; do not set to False unless you
            know what you're doing!
        infer_schema: Whether to infer the schema if missing.
        isolate_tables: Whether to isolate table selects.
        qualify_columns: Whether to qualify columns.
        allow_partial_qualification: Whether to allow partial qualification.
        validate_qualify_columns: Whether to validate columns.
        quote_identifiers: Whether to run the quote_identifiers step.
            This step is necessary to ensure correctness for case sensitive queries.
            But this flag is provided in case this step is performed at a later time.
        identify: If True, quote all identifiers, else only necessary ones.
        canonicalize_table_aliases: Whether to use canonical aliases (_0, _1, ...) for all sources
            instead of preserving table names.
        on_qualify: Callback after a table has been qualified.
        sql: Original SQL string for error highlighting. If not provided, errors will not include
            highlighting. Requires that the expression has position metadata from parsing.

    Returns:
        The qualified expression.
    """
    schema = ensure_schema(schema, dialect=dialect)
    dialect = Dialect.get_or_raise(dialect)

    expression = normalize_identifiers(
        expression,
        dialect=dialect,
        store_original_column_identifiers=True,
    )
    expression = qualify_tables(
        expression,
        db=db,
        catalog=catalog,
        dialect=dialect,
        on_qualify=on_qualify,
        canonicalize_table_aliases=canonicalize_table_aliases,
    )

    if isolate_tables:
        expression = isolate_table_selects(expression, schema=schema)

    if qualify_columns:
        expression = qualify_columns_func(
            expression,
            schema,
            expand_alias_refs=expand_alias_refs,
            expand_stars=expand_stars,
            infer_schema=infer_schema,
            allow_partial_qualification=allow_partial_qualification,
        )

    if quote_identifiers:
        expression = quote_identifiers_func(expression, dialect=dialect, identify=identify)

    if validate_qualify_columns:
        validate_qualify_columns_func(expression, sql=sql)

    return expression
